# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/000_AIOKafkaImports.ipynb.

# %% auto 0
__all__ = ['dummy']

# %% ../nbs/000_AIOKafkaImports.ipynb 1
from aiokafka import AIOKafkaConsumer, AIOKafkaProducer

from ._components.meta import export

__all__ = [
    "AIOKafkaConsumer",
    "AIOKafkaProducer",
]

# %% ../nbs/000_AIOKafkaImports.ipynb 2
@export("_dummy")
def dummy() -> None:
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/013_ProducerDecorator.ipynb.

# %% auto 0
__all__ = ['logger', 'BaseSubmodel', 'ProduceReturnTypes', 'ProduceCallable', 'KafkaEvent', 'unwrap_from_kafka_event',
           'release_callback', 'produce_single', 'send_batch', 'produce_batch', 'producer_decorator']

# %% ../../nbs/013_ProducerDecorator.ipynb 1
import asyncio
import functools
import random
import time
from asyncio import iscoroutinefunction  # do not use the version from inspect
from dataclasses import dataclass
from inspect import Parameter
from typing import *

from aiokafka import AIOKafkaProducer
from aiokafka.errors import KafkaTimeoutError
from aiokafka.producer.message_accumulator import BatchBuilder
from pydantic import BaseModel

from .logger import get_logger
from .meta import export
from .helpers import remove_suffix

# %% ../../nbs/013_ProducerDecorator.ipynb 3
logger = get_logger(__name__)

# %% ../../nbs/013_ProducerDecorator.ipynb 5
BaseSubmodel = TypeVar("BaseSubmodel", bound=Union[List[BaseModel], BaseModel])
BaseSubmodel


@dataclass
@export("fastkafka")
class KafkaEvent(Generic[BaseSubmodel]):
    """
    A generic class for representing Kafka events. Based on BaseSubmodel, bound to pydantic.BaseModel

    Attributes:
        message (BaseSubmodel): The message contained in the Kafka event, can be of type pydantic.BaseModel.
        key (bytes, optional): The optional key used to identify the Kafka event.
    """

    message: BaseSubmodel
    key: Optional[bytes] = None

# %% ../../nbs/013_ProducerDecorator.ipynb 7
def unwrap_from_kafka_event(var_type: Union[Type, Parameter]) -> Union[Type, Parameter]:
    """
    Unwraps the type from a KafkaEvent.

    Args:
        var_type: Type to unwrap.

    Returns:
        Type: Unwrapped type if the given type is a KafkaEvent, otherwise returns the same type.

    Example:
        - Input: KafkaEvent[str]
          Output: str
        - Input: int
          Output: int
    """
    if hasattr(var_type, "__origin__") and var_type.__origin__ == KafkaEvent:
        return var_type.__args__[0]  # type: ignore
    else:
        return var_type

# %% ../../nbs/013_ProducerDecorator.ipynb 9
ProduceReturnTypes = Union[
    BaseModel, KafkaEvent[BaseModel], List[BaseModel], KafkaEvent[List[BaseModel]]
]

ProduceCallable = Union[
    Callable[..., ProduceReturnTypes], Callable[..., Awaitable[ProduceReturnTypes]]
]

# %% ../../nbs/013_ProducerDecorator.ipynb 12
def _wrap_in_event(
    message: Union[BaseModel, List[BaseModel], KafkaEvent]
) -> KafkaEvent:
    return message if type(message) == KafkaEvent else KafkaEvent(message)

# %% ../../nbs/013_ProducerDecorator.ipynb 15
def release_callback(fut: asyncio.Future) -> None:
    pass

# %% ../../nbs/013_ProducerDecorator.ipynb 16
async def produce_single(  # type: ignore
    producer: AIOKafkaProducer,
    topic: str,
    encoder_fn: Callable[[BaseModel], bytes],
    wrapped_val: KafkaEvent[BaseModel],
) -> None:
    """
    Sends a single message to the Kafka producer.

    Args:
        producer (AIOKafkaProducer): The Kafka producer object.
        topic (str): The topic to which the message will be sent.
        encoder_fn (Callable[[BaseModel], bytes]): The encoding function to encode the message.
        wrapped_val (KafkaEvent[BaseModel]): The wrapped Kafka event containing the message.
    """
    while True:
        try:
            fut = await producer.send(
                topic, encoder_fn(wrapped_val.message), key=wrapped_val.key
            )
            fut.add_done_callback(release_callback)
            break
        except KafkaTimeoutError as e:
            logger.warning(
                f"produce_single(): Exception {e} raised when producing {wrapped_val.message} to {topic=}, sleeping for 1 second and retrying.."
            )
            await asyncio.sleep(1)

# %% ../../nbs/013_ProducerDecorator.ipynb 20
async def send_batch(  # type: ignore
    producer: AIOKafkaProducer, topic: str, batch: BatchBuilder, key: Optional[bytes]
) -> None:
    """
    Sends a batch of messages to the Kafka producer.

    Args:
        producer (AIOKafkaProducer): The Kafka producer object.
        topic (str): The topic to which the messages will be sent.
        batch (BatchBuilder): The batch builder object containing the messages.
        key (Optional[bytes]): The optional key used to identify the batch of messages.

    Returns:
        None
    """
    partitions = await producer.partitions_for(topic)
    if key == None:
        partition = random.choice(tuple(partitions))  # nosec
    else:
        partition = producer._partition(topic, None, None, None, key, None)
    while True:
        try:
            await producer.send_batch(batch, topic, partition=partition)
            break
        except KafkaTimeoutError as e:
            logger.warning(
                f"send_batch(): Exception {e} raised when producing {batch} to {topic=}, sleeping for 1 second and retrying.."
            )
            await asyncio.sleep(1)


async def produce_batch(  # type: ignore
    producer: AIOKafkaProducer,
    topic: str,
    encoder_fn: Callable[[BaseModel], bytes],
    wrapped_val: KafkaEvent[List[BaseModel]],
) -> ProduceReturnTypes:
    """
    Sends a batch of messages to the Kafka producer.

    Args:
        producer (AIOKafkaProducer): The Kafka producer object.
        topic (str): The topic to which the messages will be sent.
        encoder_fn (Callable[[BaseModel], bytes]): The encoding function to encode the messages.
        wrapped_val (KafkaEvent[List[BaseModel]]): The wrapped Kafka event containing the list of messages.

    Returns:
        ProduceReturnTypes: The return value from the decorated function.
    """
    batch = producer.create_batch()

    for message in wrapped_val.message:
        metadata = batch.append(
            key=wrapped_val.key,
            value=encoder_fn(message),
            timestamp=int(time.time() * 1000),
        )
        if metadata == None:
            # send batch
            await send_batch(producer, topic, batch, wrapped_val.key)
            # create new batch
            batch = producer.create_batch()
            batch.append(
                key=None, value=encoder_fn(message), timestamp=int(time.time() * 1000)
            )

    await send_batch(producer, topic, batch, wrapped_val.key)

# %% ../../nbs/013_ProducerDecorator.ipynb 23
def producer_decorator(
    producer_store: Dict[str, Any],
    func: ProduceCallable,
    topic_key: str,
    encoder_fn: Callable[[BaseModel], bytes],
) -> ProduceCallable:
    """
    Decorator for Kafka producer functions.

    Args:
        producer_store (Dict[str, Any]): Dictionary to store the Kafka producer objects.
        func (ProduceCallable): The function to be decorated.
        topic_key (str): The key used to identify the topic.
        encoder_fn (Callable[[BaseModel], bytes]): The encoding function to encode the messages.

    Returns:
        ProduceCallable: The decorated function.

    Raises:
        ValueError: If the decorated function is synchronous.
    """

    @functools.wraps(func)
    async def _produce_async(
        *args: List[Any],
        topic_key: str = topic_key,
        encoder_fn: Callable[[BaseModel], bytes] = encoder_fn,
        producer_store: Dict[str, Any] = producer_store,
        f: Callable[..., Awaitable[ProduceReturnTypes]] = func,  # type: ignore
        **kwargs: Any,
    ) -> ProduceReturnTypes:
        return_val = await f(*args, **kwargs)
        wrapped_val = _wrap_in_event(return_val)
        _, producer, _, _ = producer_store[topic_key]
        topic = remove_suffix(topic_key)

        if isinstance(wrapped_val.message, list):
            await produce_batch(producer, topic, encoder_fn, wrapped_val)
        else:
            await produce_single(producer, topic, encoder_fn, wrapped_val)
        return return_val

    if not iscoroutinefunction(func):
        raise ValueError(
            "Synchronous functions are not supported for produce operation"
        )

    return _produce_async



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/019_Json_Encode_Decoder.ipynb.

# %% auto 0
__all__ = ['logger', 'json_encoder', 'json_decoder']

# %% ../../../nbs/019_Json_Encode_Decoder.ipynb 1
import json
from typing import *

from pydantic import BaseModel

from ..logger import get_logger
from ..meta import export

# %% ../../../nbs/019_Json_Encode_Decoder.ipynb 4
logger = get_logger(__name__)

# %% ../../../nbs/019_Json_Encode_Decoder.ipynb 6
def _to_json_utf8(o: Any) -> bytes:
    """Converts to JSON and then encodes with UTF-8"""
    if hasattr(o, "model_dump_json"):
        return o.model_dump_json().encode("utf-8")  # type: ignore
    else:
        return json.dumps(o).encode("utf-8")

# %% ../../../nbs/019_Json_Encode_Decoder.ipynb 9
@export("fastkafka.encoder")
def json_encoder(msg: BaseModel) -> bytes:
    """
    Encoder to encode pydantic instances to json string

    Args:
        msg: An instance of pydantic basemodel

    Returns:
        Json string in bytes which is encoded from pydantic basemodel
    """
    return _to_json_utf8(msg)

# %% ../../../nbs/019_Json_Encode_Decoder.ipynb 11
@export("fastkafka.encoder")
def json_decoder(raw_msg: bytes, cls: Type[BaseModel]) -> Any:
    """
    Decoder to decode json string in bytes to pydantic model instance

    Args:
        raw_msg: Bytes message received from Kafka topic
        cls: Pydantic class; This pydantic class will be used to construct instance of same class

    Returns:
        An instance of given pydantic class
    """
    msg_dict = json.loads(raw_msg.decode("utf-8"))

    return cls(**msg_dict)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/005_Application_executors_export.ipynb.

# %% auto 0
__all__ = ['dummy']

# %% ../nbs/005_Application_executors_export.ipynb 1
from ._components.meta import export
from ._components.task_streaming import SequentialExecutor, DynamicTaskExecutor

__all__ = ["SequentialExecutor", "DynamicTaskExecutor"]

# %% ../nbs/005_Application_executors_export.ipynb 2
@export("_dummy")
def dummy() -> None:
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/998_Internal_Helpers.ipynb.

# %% auto 0
__all__ = ['in_notebook', 'change_dir', 'ImportFromStringError', 'true_after', 'unwrap_list_type', 'remove_suffix']

# %% ../../nbs/998_Internal_Helpers.ipynb 2
def in_notebook() -> bool:
    """
    Checks if the code is running in a Jupyter notebook or not.

    Returns:
        True if running in a Jupyter notebook, False otherwise.
    """
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True

# %% ../../nbs/998_Internal_Helpers.ipynb 4
import contextlib
import importlib
import os
import sys
from datetime import datetime, timedelta
from inspect import Parameter
from typing import *

import typer

# %% ../../nbs/998_Internal_Helpers.ipynb 6
@contextlib.contextmanager
def change_dir(d: str) -> Generator[None, None, None]:
    """
    Changes the current working directory temporarily.

    Args:
        d: The directory to change to.

    Yields:
        None.
    """
    curdir = os.getcwd()
    os.chdir(d)
    try:
        yield
    finally:
        os.chdir(curdir)

# %% ../../nbs/998_Internal_Helpers.ipynb 8
class ImportFromStringError(Exception):
    pass


def _import_from_string(import_str: str) -> Any:
    """Imports library from string

    Note:
        copied from https://github.com/encode/uvicorn/blob/master/uvicorn/importer.py

    Args:
        import_str: input string in form 'main:app'

    """
    sys.path.append(".")

    if not isinstance(import_str, str):
        return import_str

    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = (
            'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )
        typer.secho(f"{message}", err=True, fg=typer.colors.RED)
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        # nosemgrep: python.lang.security.audit.non-literal-import.non-literal-import
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(
            message.format(attrs_str=attrs_str, module_str=module_str)
        )

    return instance

# %% ../../nbs/998_Internal_Helpers.ipynb 10
def true_after(seconds: Union[int, float]) -> Callable[[], bool]:
    """Function returning True after a given number of seconds"""
    t = datetime.now()

    def _true_after(seconds: Union[int, float] = seconds, t: datetime = t) -> bool:
        return (datetime.now() - t) > timedelta(seconds=seconds)

    return _true_after

# %% ../../nbs/998_Internal_Helpers.ipynb 12
def unwrap_list_type(var_type: Union[Type, Parameter]) -> Union[Type, Parameter]:
    """
    Unwraps the type of a list.

    Vars:
        var_type: Type to unwrap.

    Returns:
        Unwrapped type if the given type is a list, otherwise returns the same type.

    Example:
        - Input: List[str]
          Output: str
        - Input: int
          Output: int
    """
    if hasattr(var_type, "__origin__") and var_type.__origin__ == list:
        return var_type.__args__[0]  # type: ignore
    else:
        return var_type

# %% ../../nbs/998_Internal_Helpers.ipynb 14
def remove_suffix(topic: str) -> str:
    """
    Removes the suffix from a string by splitting on underscores and joining all but the last element.

    Args:
        topic: The string to remove the suffix from.

    Returns:
        The string without the suffix.
    """
    return "_".join(topic.split("_")[:-1])



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/096_Docusaurus_Helper.ipynb.

# %% auto 0
__all__ = ['CustomNbdevLookup', 'fix_invalid_syntax_in_markdown', 'generate_markdown_docs', 'generate_sidebar',
           'delete_unused_markdown_files_from_sidebar', 'update_readme']

# %% ../nbs/096_Docusaurus_Helper.ipynb 2
import itertools
import re
import ast
import types
from inspect import (
    Signature,
    getmembers,
    isclass,
    isfunction,
    signature,
    ismethod,
    getsource,
    Parameter,
)
from pathlib import Path
from typing import *
from urllib.parse import urljoin
from functools import lru_cache

import typer
from docstring_parser import parse
from docstring_parser.common import (
    DocstringParam,
    DocstringRaises,
    DocstringReturns,
    Docstring,
)
from nbdev.config import get_config
from nbdev.quarto import nbdev_readme
from nbdev.doclinks import NbdevLookup, patch_name, L, _find_mod
from nbdev_mkdocs.mkdocs import (
    _add_all_submodules,
    _import_all_members,
    _import_functions_and_classes,
    _import_submodules,
)
from nbdev_mkdocs._helpers.doc_links_utils import (
    fix_sym_links as update_default_symbol_links,
)

# %% ../nbs/096_Docusaurus_Helper.ipynb 4
def _get_return_annotation(s: Signature) -> str:
    """Get the return annotation from the function signature.

    Args:
        s: The signature of the function from which the annotations must be extracted.

    Returns:
        The return annotation, or an empty string if not available.

    """
    if s.return_annotation == None or "inspect._empty" in str(s.return_annotation):
        return ""
    if isinstance(s.return_annotation, str):
        return s.return_annotation
    ret_val: str = (
        str(s.return_annotation).replace("typing.", "").replace("NoneType", "None")
        if "typing." in str(s.return_annotation)
        else str(s.return_annotation.__name__)
    )
    return ret_val

# %% ../nbs/096_Docusaurus_Helper.ipynb 19
def _get_param_annotation(param: Parameter) -> str:
    """Get the annotation of a function parameter.

    Args:
        param: The parameter object.

    Returns:
        The parameter annotation, or an empty string if not available.

    """

    if "typing." in str(param.annotation):
        return f'`{str(param.annotation).replace("typing.", "")}`'
    elif isinstance(param.annotation, str):
        return param.annotation
    else:
        return (
            ""
            if param.annotation.__name__ == "_empty"
            else f"`{param.annotation.__name__}`"
        )

# %% ../nbs/096_Docusaurus_Helper.ipynb 21
def _get_default_value(param: Parameter) -> str:
    """Get the default value of the function parameter.

    Args:
        param: The parameter object.

    Returns:
        The default value of the function parameter.

    """
    if param.default is param.empty:
        return "*required*"

    return (
        f"`'{param.default}'`"
        if isinstance(param.default, str)
        else f"`{param.default}`"
    )

# %% ../nbs/096_Docusaurus_Helper.ipynb 23
def _get_params_annotation(s: Signature) -> Dict[str, Dict[str, str]]:
    """Get the annotations along with its default values for the parameters of the symbol.

    Args:
        s: The signature of the function from which the annotations must be extracted.

    Returns:
        The parameter annotations along with its default value.
    """
    return {
        f"{param.name}": {
            "type": _get_param_annotation(param),
            "default": _get_default_value(param),
        }
        for param in s.parameters.values()
    }

# %% ../nbs/096_Docusaurus_Helper.ipynb 25
def _generate_parameters_table(
    symbol_annotations: Dict[str, Union[Dict[str, str], str]],
    section_items: Union[List[DocstringParam]],
    section_name: str,
) -> str:
    """Generate parameter table in markdown format

    Args:
        symbol_annotations: Symbol annotations along with its default value
        section_items: The parameter section of a parsed docstring
        section_name: The name of the section

    Returns:
        The parameters of a symbol in markdown-formatted string
    """
    nl = "\n"
    _section_template = (
        "|  Name | Type | Description | Default |\n|---|---|---|---|\n{section_body}\n"
    )
    section_body = "".join(
        [
            f'| `{section.arg_name}` | {symbol_annotations["parameters"][section.arg_name]["type"]} | {section.description.replace(nl, "")} | {symbol_annotations["parameters"][section.arg_name]["default"]} |\n'  # type: ignore
            if section.arg_name in symbol_annotations["parameters"]
            else ""
            for section in section_items
        ]
    )
    return f"**{section_name}**:\n\n" + _section_template.format(
        section_body=section_body,
    )

# %% ../nbs/096_Docusaurus_Helper.ipynb 27
def _generate_return_and_raises_table(
    symbol_annotations: Dict[str, Union[Dict[str, str], str]],
    section_items: Union[List[DocstringReturns], List[DocstringRaises]],
    section_name: str,
) -> str:
    """Generate return and raises table in markdown format

    Args:
        symbol_annotations: Symbol annotations along with its default value
        section_items: The parameter section of a parsed docstring
        section_name: The name of the section

    Returns:
        The return and raises section of a symbol in markdown-formatted string
    """
    nl = "\n"
    _section_template = "|  Type | Description |\n|---|---|\n{section_body}\n"
    section_body = "".join(
        [
            f'| `{symbol_annotations["return"] if section_name == "Returns" else section.type_name}` | {section.description.replace(nl, "")} |\n'  # type: ignore
            for section in section_items
        ]
    )
    return f"**{section_name}**:\n\n" + _section_template.format(
        section_body=section_body,
    )

# %% ../nbs/096_Docusaurus_Helper.ipynb 30
def _format_docstring_section_items(
    symbol_annotations: Dict[str, Union[Dict[str, str], str]],
    section_items: Union[
        List[DocstringParam], List[DocstringReturns], List[DocstringRaises]
    ],
    section_name: str,
) -> str:
    """Format the docstring sections in a table format

    Args:
        symbol_annotations: Symbol annotations along with its default value
        section_items: The parameter section of a parsed docstring
        section_name: The name of the section

    Returns:
        The docstring sections of the symbol in markdown-formatted string
    """
    if section_name == "Parameters":
        return _generate_parameters_table(symbol_annotations, section_items, section_name)  # type: ignore
    else:
        return _generate_return_and_raises_table(symbol_annotations, section_items, section_name)  # type: ignore

# %% ../nbs/096_Docusaurus_Helper.ipynb 35
def _get_annotation(symbol: Type) -> Dict[str, Union[Dict[str, Dict[str, str]], str]]:
    """Get annotations along with its default value for a symbol

    Args:
        symbol: The symbol for which the annotations needs to be extracted

    Returns:
        The annotations dict along with its default value
    """
    symbol = symbol.fget if isinstance(symbol, property) else symbol
    symbol_signature = signature(symbol)
    params_dict = _get_params_annotation(symbol_signature)
    return_annotation = _get_return_annotation(symbol_signature)
    return {"parameters": params_dict, "return": return_annotation}

# %% ../nbs/096_Docusaurus_Helper.ipynb 38
def _format_docstring_sections(symbol: Type, parsed_docstring: Docstring) -> str:
    """Format the parsed docstring sections into markdown-formatted table

    Args:
        symbol: The symbol for which to parse the docstring.
        parsed_docstring: A Docstring object

    Returns:
        The markdown-formatted docstring.
    """
    symbol_annotations = _get_annotation(symbol)
    formatted_docstring = ""
    sections = [
        ("Parameters", parsed_docstring.params),
        ("Returns", parsed_docstring.many_returns),
        ("Exceptions", parsed_docstring.raises),
    ]

    for section_name, section_items in sections:
        if len(section_items) > 0:  # type: ignore
            formatted_docstring += _format_docstring_section_items(
                symbol_annotations, section_items, section_name  # type: ignore
            )

    return formatted_docstring

# %% ../nbs/096_Docusaurus_Helper.ipynb 40
def _format_free_links(s: str) -> str:
    """Format free links in a given string by adding proper spacing around them.

    Args:
        s: The input string containing free links.

    Returns:
        The modified string with properly formatted free links.
    """
    pattern = r"([\"'])(https?:\/\/[^\s]+)([\"'])"
    ret_val = re.sub(
        pattern, lambda match: f"{match.group(1)} {match.group(2)} {match.group(3)}", s
    )
    return ret_val

# %% ../nbs/096_Docusaurus_Helper.ipynb 42
def _docstring_to_markdown(symbol: Type) -> str:
    """Converts a docstring to a markdown-formatted string.

    Args:
        symbol: The symbol for which the documentation needs to be generated in markdown format.

    Returns:
        The markdown-formatted docstring.
    """
    parsed_docstring = parse(symbol.__doc__)  # type: ignore
    formatted_docstring = f"{parsed_docstring.short_description}\n\n"
    formatted_docstring += (
        f"{parsed_docstring.long_description}\n\n"
        if parsed_docstring.long_description
        else ""
    )
    formatted_docstring += _format_docstring_sections(symbol, parsed_docstring)
    ret_val = _format_free_links(formatted_docstring)

    return ret_val

# %% ../nbs/096_Docusaurus_Helper.ipynb 46
def _get_submodules(module_name: str) -> List[str]:
    """Get a list of all submodules contained within the module.

    Args:
        module_name: The name of the module to retrieve submodules from

    Returns:
        A list of submodule names within the module
    """
    members = _import_all_members(module_name)
    members_with_submodules = _add_all_submodules(members)
    members_with_submodules_str: List[str] = [
        x[:-1] if x.endswith(".") else x for x in members_with_submodules
    ]
    return members_with_submodules_str

# %% ../nbs/096_Docusaurus_Helper.ipynb 48
def _load_submodules(
    module_name: str, members_with_submodules: List[str]
) -> List[Type]:
    """Load the given submodules from the module.

    Args:
        module_name: The name of the module whose submodules to load
        members_with_submodules: A list of submodule names to load

    Returns:
        A list of imported submodule objects.
    """
    submodules = _import_submodules(module_name)
    members: List[Tuple[str, Type]] = list(
        itertools.chain(*[_import_functions_and_classes(m) for m in submodules])
    )
    names = [
        y
        for x, y in members
        if f"{y.__module__}.{y.__name__}" in members_with_submodules
    ]
    return names

# %% ../nbs/096_Docusaurus_Helper.ipynb 50
def _get_parameters(_signature: Signature) -> List[str]:
    """Convert a function's signature into a string representation of its parameter list.

    Args:
        _signature: The signature object representing the function's signature.

    Returns:
        A list of strings representing the function's parameters, including their default values if applicable.
    """
    params = [param for param in _signature.parameters.values()]
    ret_val = [
        f"{param.name}"
        if (param.default is param.empty)
        else f"{param.name}='{param.default}'"
        if isinstance(param.default, str)
        else f"{param.name}={param.default}"
        for param in params
    ]
    return ret_val

# %% ../nbs/096_Docusaurus_Helper.ipynb 54
def _format_symbol_definition(symbol: Type, params_list: List[str]) -> str:
    """Format the given symbol parameters by adding a new line and indentation.

    Args:
        symbol: The symbol for which the symbol definition needs to be formatted.
        params_list: A string representation of the parameter list.

    Returns:
        A formatted string representation of the parameters with new lines and indentation.
    """
    parameters = ", ".join(params_list)
    if parameters == "":
        return f"{symbol.__name__}()\n"
    elif len(f"{symbol.__name__}({parameters})") <= 79:
        return f"{symbol.__name__}(\n    {parameters}\n)\n"
    else:
        formatted_parameters = "".join([f"\n    {param}," for param in params_list])
        return f"{symbol.__name__}({formatted_parameters}\n)\n"

# %% ../nbs/096_Docusaurus_Helper.ipynb 61
def _get_exps(mod: str) -> Dict[str, str]:
    mf = _find_mod(mod)
    if not mf:
        return {}
    txt = mf.read_text()
    _def_types = ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef
    d = {}
    for tree in ast.parse(txt).body:
        if isinstance(tree, _def_types):
            for t in L(patch_name(tree)):
                d[t] = f"{tree.lineno}-L{tree.end_lineno}"
        if isinstance(tree, ast.ClassDef):
            d.update(
                {
                    tree.name + "." + t2.name: f"{t2.lineno}-L{t2.end_lineno}"
                    for t2 in tree.body
                    if isinstance(t2, _def_types)
                }
            )
    return d

# %% ../nbs/096_Docusaurus_Helper.ipynb 63
def _lineno(sym: str, fname: str) -> Optional[str]:
    return _get_exps(fname).get(sym, None) if fname else None


@lru_cache(None)
class CustomNbdevLookup(NbdevLookup.__wrapped__):  # type: ignore
    def __init__(
        self,
        strip_libs: Optional[str] = None,
        incl_libs: Optional[str] = None,
        skip_mods: Optional[str] = None,
    ):
        super().__init__(strip_libs, incl_libs, skip_mods)

    def code(self, sym: str) -> Optional[str]:
        "Link to source code for `sym`"
        res = self[sym]
        if not isinstance(res, tuple):
            return None
        _, py, gh = res
        line = _lineno(sym, py)
        return f"{gh}#L{line}"

# %% ../nbs/096_Docusaurus_Helper.ipynb 66
def _get_symbol_source_link(symbol: Type, lib_version: str) -> str:
    """Returns the source code link for a given symbol.

    Args:
        symbol: The symbol to get the source code link for.
        lib_version: The current version of the library.

    Returns:
        The source code link for the symbol.
    """
    symbol = symbol.fget if isinstance(symbol, property) else symbol
    source_link = CustomNbdevLookup().code(f"{symbol.__qualname__}")

    if source_link is None:
        return ""

    href = (
        source_link.replace("/blob/main/", f"/blob/{lib_version}/")
        if lib_version.replace(".", "").isdigit()
        else source_link
    )
    return f'<a href="{href}" class="link-to-source" target="_blank">View source</a>'

# %% ../nbs/096_Docusaurus_Helper.ipynb 70
def _get_method_type(symbol: Type) -> str:
    try:
        source = getsource(symbol).strip()
    except TypeError as e:
        return ""

    first_line = source.split("\n")[0]
    return (
        f"{first_line}\n"
        if first_line
        in ["@abstractmethod", "@staticmethod", "@classmethod", "@property"]
        else ""
    )


def _get_symbol_definition(symbol: Type, header_level: int, lib_version: str) -> str:
    """Return the definition of a given symbol.

    Args:
        symbol: A function or method object to get the definition for.
        header_level: The level of the markdown header to append.
        lib_version: The current version of the library.

    Returns:
        A string representing the function definition
    """
    if isclass(symbol):
        return f"{'#'*(header_level - 1)} {symbol.__module__}.{symbol.__name__} {{#{symbol.__module__}.{symbol.__name__}}}\n\n{_get_symbol_source_link(symbol, lib_version)}\n\n"

    if isinstance(symbol, property):
        symbol = symbol.fget

    symbol_anchor = (
        f"{'#' * header_level} {symbol.__name__}"
        + f" {{#{symbol.__module__}.{'.'.join([component.strip('_') for component in symbol.__qualname__.rsplit('.', 1)])}}}\n\n"
    )

    link_to_source = f"{_get_symbol_source_link(symbol, lib_version)}\n\n"

    _signature = signature(symbol)
    parameters = _get_parameters(_signature)
    symbol_definition = f"```py\n{_get_method_type(symbol)}{_format_symbol_definition(symbol, parameters)}```\n"
    return symbol_anchor + link_to_source + symbol_definition

# %% ../nbs/096_Docusaurus_Helper.ipynb 76
def _is_method(symbol: Type) -> bool:
    """Check if the given symbol is a method.

    Args:
        symbol: A function or method object to check.

    Returns:
        A boolean indicating whether the symbol is a method.
    """
    return ismethod(symbol) or isfunction(symbol) or isinstance(symbol, property)

# %% ../nbs/096_Docusaurus_Helper.ipynb 78
def _get_formatted_docstring_for_symbol(
    symbol: Type, lib_version: str, header_level: int = 2
) -> str:
    """Recursively parses and get formatted docstring of a symbol.

    Args:
        symbol: A Python class or function object to parse the docstring for.
        lib_version: The current version of the library.
        header_level: The level of the markdown header to append.

    Returns:
        A formatted docstring of the symbol and its members.

    """

    def traverse(
        symbol: Type, contents: str, header_level: int, lib_version: str
    ) -> str:
        """Recursively traverse the members of a symbol and append their docstrings to the provided contents string.

        Args:
            symbol: A Python class or function object to parse the docstring for.
            contents: The current formatted docstrings.
            header_level: The level of the markdown header to append.
            lib_version: The current version of the library.

        Returns:
            The updated formatted docstrings.

        """
        for x, y in getmembers(symbol):
            if not x.startswith("_") or x == "__init__":
                if _is_method(y) and y.__doc__ is not None:
                    contents += f"{_get_symbol_definition(y, header_level, lib_version)}\n{_docstring_to_markdown(y)}"
                elif isclass(y) and y.__doc__ is not None and not x.startswith("_"):
                    contents += f"{_get_symbol_definition(y, header_level+1, lib_version)}\n{_docstring_to_markdown(y)}"
                    contents = traverse(y, contents, header_level + 1, lib_version)
        return contents

    contents = (
        f"{_get_symbol_definition(symbol, header_level+1, lib_version)}\n{_docstring_to_markdown(symbol)}"
        if symbol.__doc__ is not None
        else ""
    )
    if isclass(symbol):
        contents = traverse(symbol, contents, header_level + 1, lib_version)
    return contents

# %% ../nbs/096_Docusaurus_Helper.ipynb 81
def _convert_html_style_attribute_to_jsx(contents: str) -> str:
    """Converts the inline style attributes in an HTML string to JSX compatible format.

    Args:
        contents: A string containing an HTML document or fragment.

    Returns:
        A string with inline style attributes converted to JSX compatible format.
    """
    style_regex = re.compile(r'style="(.+?)"')
    style_matches = style_regex.findall(contents)

    for style_match in style_matches:
        style_dict = {}
        styles = style_match.split(";")
        for style in styles:
            key_value = style.split(":")
            if len(key_value) == 2:
                key = re.sub(
                    r"-(.)", lambda m: m.group(1).upper(), key_value[0].strip()
                )
                value = key_value[1].strip().replace("'", '"')
                style_dict[key] = value
        replacement = "style={{"
        for key, value in style_dict.items():
            replacement += f"{key}: '{value}', "
        replacement = replacement[:-2] + "}}"
        contents = contents.replace(f'style="{style_match}"', replacement)

    return contents

# %% ../nbs/096_Docusaurus_Helper.ipynb 83
def _get_all_markdown_files_path(docs_path: Path) -> List[Path]:
    """Get all Markdown files in a directory and its subdirectories.

    Args:
        directory: The path to the directory to search in.

    Returns:
        A list of paths to all Markdown files found in the directory and its subdirectories.
    """
    markdown_files = [file_path for file_path in docs_path.glob("**/*.md")]
    return markdown_files

# %% ../nbs/096_Docusaurus_Helper.ipynb 85
def _fix_special_symbols_in_html(contents: str) -> str:
    contents = contents.replace("”", '"')
    return contents

# %% ../nbs/096_Docusaurus_Helper.ipynb 87
def _add_file_extension_to_link(url: str) -> str:
    """Add file extension to the last segment of a URL

    Args:
        url: A URL string.

    Returns:
        A string of the updated URL with a file extension added to the last segment of the URL.
    """
    segments = url.split("/#")[0].split("/")[-2:]
    return url.replace(f"/{segments[1]}", f"/{segments[1]}.md").replace(".md/#", ".md#")

# %% ../nbs/096_Docusaurus_Helper.ipynb 91
def _generate_production_url(url: str) -> str:
    """Generate a Docusaurus compatible production URL for the given symbol URL.

    Args:
        url: The symbol URL to be converted.

    Returns:
        The production URL of the symbol.
    """
    url_segment, hash_segment = url.split(".md")
    url_split = url_segment.split("/")
    if url_split[-1].lower() == url_split[-2].lower():
        return "/".join(url_split[:-1]) + hash_segment
    return url.replace(".md", "")

# %% ../nbs/096_Docusaurus_Helper.ipynb 93
def _fix_symbol_links(
    contents: str,
    dir_prefix: str,
    doc_host: str,
    doc_baseurl: str,
    use_relative_doc_links: bool = True,
) -> str:
    """Fix symbol links in Markdown content.

    Args:
        contents: The Markdown content to search for symbol links.
        dir_prefix: Directory prefix to append in the relative URL.
        doc_host: The host URL for the documentation site.
        doc_baseurl: The base URL for the documentation site.
        use_relative_doc_links: If set to True, then the relative link to symbols will be added else,
            production link will be added.

    Returns:
        str: The Markdown content with updated symbol links.
    """
    prefix = re.escape(urljoin(doc_host + "/", doc_baseurl))
    pattern = re.compile(rf"\[(.*?)\]\(({prefix}[^)]+)\)")
    matches = pattern.findall(contents)
    for match in matches:
        old_url = match[1]
        new_url = _add_file_extension_to_link(old_url).replace("/api/", "/docs/api/")
        if use_relative_doc_links:
            dir_prefix = "./" if dir_prefix == "" else dir_prefix
            updated_url = dir_prefix + new_url.split("/docs/")[1]
        else:
            updated_url = _generate_production_url(
                doc_host + doc_baseurl + "/docs/" + new_url.split("/docs/")[1]
            )
        contents = contents.replace(old_url, updated_url)
    return contents

# %% ../nbs/096_Docusaurus_Helper.ipynb 101
def _get_relative_url_prefix(docs_path: Path, sub_path: Path) -> str:
    """Returns a relative url prefix from a sub path to a docs path.

    Args:
        docs_path (Path): The docs directory path.
        sub_path (Path): The sub directory path.

    Returns:
        str: A string representing the relative path from the sub path to the docs path.

    Raises:
        ValueError: If the sub path is not a descendant of the docs path.
    """
    try:
        relative_path = sub_path.relative_to(docs_path)
    except ValueError:
        raise ValueError(f"{sub_path} is not a descendant of {docs_path}")

    return (
        "../" * (len(relative_path.parts) - 1) if len(relative_path.parts) > 1 else ""
    )

# %% ../nbs/096_Docusaurus_Helper.ipynb 103
def fix_invalid_syntax_in_markdown(docs_path: str) -> None:
    """Fix invalid HTML syntax in markdown files and converts inline style attributes to JSX-compatible format.

    Args:
        docs_path: The path to the root directory to search for markdown files.
    """
    cfg = get_config()
    doc_host = cfg["doc_host"]
    doc_baseurl = cfg["doc_baseurl"]

    markdown_files = _get_all_markdown_files_path(Path(docs_path))
    for file in markdown_files:
        relative_url_prefix = _get_relative_url_prefix(Path(docs_path), file)
        contents = Path(file).read_text()

        contents = _convert_html_style_attribute_to_jsx(contents)
        contents = _fix_special_symbols_in_html(contents)
        contents = _fix_symbol_links(
            contents, relative_url_prefix, doc_host, doc_baseurl
        )
        file.write_text(contents)

# %% ../nbs/096_Docusaurus_Helper.ipynb 105
def generate_markdown_docs(module_name: str, docs_path: str) -> None:
    """Generates Markdown documentation files for the symbols in the given module and save them to the given directory.

    Args:
        module_name: The name of the module to generate documentation for.
        docs_path: The path to the directory where the documentation files will be saved.
    """
    members_with_submodules = _get_submodules(module_name)
    symbols = _load_submodules(module_name, members_with_submodules)
    lib_version = get_config()["version"]

    for symbol in symbols:
        content = _get_formatted_docstring_for_symbol(symbol, lib_version)
        target_file_path = (
            "/".join(f"{symbol.__module__}.{symbol.__name__}".split(".")) + ".md"
        )
        with open((Path(docs_path) / "api" / target_file_path), "w") as f:
            f.write(content)

# %% ../nbs/096_Docusaurus_Helper.ipynb 108
def _parse_lines(lines: List[str]) -> Tuple[List[str], int]:
    """Parse a list of lines and return a tuple containing a list of filenames and an index indicating how many lines to skip.

    Args:
        lines: A list of strings representing lines of input text.

    Returns:
        A tuple containing a list of strings representing the filenames extracted
        from links in the lines and an integer representing the number of lines to skip.
    """
    index = next(
        (i for i, line in enumerate(lines) if not line.strip().startswith("- [")),
        len(lines),
    )
    return [line.split("(")[1][:-4] for line in lines[:index]], index

# %% ../nbs/096_Docusaurus_Helper.ipynb 111
def _parse_section(text: str, ignore_first_line: bool = False) -> List[Any]:
    """Parse the given section contents and return a list of file names in the expected format.

    Args:
        text: A string representing the contents of a file.
        ignore_first_line: Flag indicating whether to ignore the first line extracting the section contents.

    Returns:
        A list of filenames in the expected format
    """
    pattern = r"\[.*?\]\((.*?)\)|\[(.*?)\]\[(.*?)\]"
    lines = text.split("\n")[1:] if ignore_first_line else text.split("\n")
    ret_val = []
    index = 0
    while index < len(lines):
        line = lines[index]
        match = re.search(pattern, line.strip())
        if match is not None:
            ret_val.append(match.group(1).split(".md")[0])
            index += 1
        elif line.strip() != "":
            value, skip_lines = _parse_lines(lines[index + 1 :])
            ret_val.append({line.replace("-", "").strip(): value})
            index += skip_lines + 1
        else:
            index += 1
    return ret_val

# %% ../nbs/096_Docusaurus_Helper.ipynb 114
def _get_section_from_markdown(
    markdown_text: str, section_header: str
) -> Optional[str]:
    """Get the contents of the section header from the given markdown text

    Args:
        markdown_text: A string containing the markdown text to extract the section from.
        section_header: A string representing the header of the section to extract.

    Returns:
        A string representing the contents of the section header if the section header
        is present in the markdown text, else None
    """
    pattern = re.compile(rf"^- {section_header}\n((?:\s+- .*\n)+)", re.M)
    match = pattern.search(markdown_text)
    return match.group(1) if match else None

# %% ../nbs/096_Docusaurus_Helper.ipynb 119
def generate_sidebar(
    summary_file: str = "./docusaurus/docs/SUMMARY.md",
    summary: str = "",
    target: str = "./docusaurus/sidebars.js",
) -> None:
    """
    Generate a sidebar js file for a Docusaurus documentation site based on a SUMMARY.md file.

    Args:
        summary_file: The path to the SUMMARY.md file containing the documentation structure.
            Default is "./docusaurus/docs/SUMMARY.md".
        summary: An optional summary string.
            Default is an empty string.
        target: The path to the target sidebar js file to be generated.
            Default is "./docusaurus/sidebars.js".

    Returns:
        None: The function does not return any value directly, but it generates a sidebar file.

    Raises:
        FileNotFoundError: If the specified `summary_file` does not exist.
    """
    with open(summary_file, "r") as stream, open(target, "w") as target_stream:
        summary_contents = stream.read()

        guides_summary = _get_section_from_markdown(summary_contents, "Guides")
        parsed_guides = _parse_section(guides_summary)  # type: ignore

        api_summary = _get_section_from_markdown(summary_contents, "API")
        parsed_api = _parse_section(api_summary, True)  # type: ignore

        cli_summary = _get_section_from_markdown(summary_contents, "CLI")
        parsed_cli = _parse_section(cli_summary)  # type: ignore

        target_stream.write(
            """module.exports = {
tutorialSidebar: [
    'index', {'Guides': 
    """
            + str(parsed_guides)
            + "},"
            + "{'API': ["
            + str(parsed_api)[1:-1]
            + "]},"
            + "{'CLI': "
            + str(parsed_cli)
            + "},"
            + """
    "LICENSE",
    "CONTRIBUTING",
    "CHANGELOG",
],
};"""
        )

# %% ../nbs/096_Docusaurus_Helper.ipynb 121
def _get_markdown_filenames_from_sidebar(sidebar_file_path: str) -> List[str]:
    """Get a list of Markdown filenames included in the sidebar.

    Args:
        sidebar_file_path: The path to the sidebar file.

    Returns:
        A list of Markdown filenames included in the sidebar.
    """
    with open(sidebar_file_path, "r") as file:
        file_content = file.read()

        pattern = r"tutorialSidebar:\s*(\[.*\])\s*,\s*\n?\s*};"
        match = re.search(pattern, file_content, re.DOTALL)
        all_sidebar_files = ast.literal_eval(match.group(1)) if match else []
        markdown_filenames = [
            f"{v}.md" for v in all_sidebar_files if isinstance(v, str)
        ]
        return markdown_filenames

# %% ../nbs/096_Docusaurus_Helper.ipynb 123
def _delete_files(files: List[Path]) -> None:
    """Deletes a list of files.

    Args:
        files: A list of Path objects representing the files to be deleted.

    Raises:
        OSError: If an error occurs while deleting a file.

    """
    for file in files:
        try:
            file.unlink()
        except OSError as e:
            typer.echo(
                f"Error deleting files from docusaurus/docs directory. Could not delete file: {file} - {e}"
            )

# %% ../nbs/096_Docusaurus_Helper.ipynb 126
def delete_unused_markdown_files_from_sidebar(
    docs_path: str, sidebar_file_path: str
) -> None:
    """Delete the markdown files from the docs directory that are not present in the sidebar.

    Args:
        docs_path: Path to the directory containing the markdown files.
        sidebar_file_path: Path to the sidebar file.
    """
    md_filenames_in_sidebar = _get_markdown_filenames_from_sidebar(
        str(sidebar_file_path)
    )
    if len(md_filenames_in_sidebar) > 0:
        all_md_files_in_docs_dir = [
            file_path for file_path in Path(docs_path).glob("*.md")
        ]
        md_files_in_sidebar = [Path(docs_path) / f for f in md_filenames_in_sidebar]
        md_files_to_delete = list(
            set(all_md_files_in_docs_dir) - set(md_files_in_sidebar)
        )
        _delete_files(md_files_to_delete)

# %% ../nbs/096_Docusaurus_Helper.ipynb 128
def update_readme() -> None:
    """Update the readme file and fix the symbol links"""
    cfg = get_config()
    readme_path = cfg.config_path / "README.md"
    nbdev_readme.__wrapped__()

    with open(readme_path, "r", encoding="utf-8") as f:
        contents = f.read()

    contents = update_default_symbol_links(
        contents, NbdevLookup(incl_libs=cfg.lib_path.name), "", "", False
    )
    contents = _fix_symbol_links(contents, "./", cfg.doc_host, cfg.doc_baseurl, False)

    with open(readme_path, "w", encoding="utf-8") as f:
        f.write(contents)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/016_Tester.ipynb.

# %% auto 0
__all__ = ['Tester', 'mirror_producer', 'mirror_consumer', 'AmbiguousWarning', 'set_sugar']

# %% ../../nbs/016_Tester.ipynb 1
import asyncio
import collections
import inspect
from unittest.mock import AsyncMock, MagicMock
import json
from contextlib import asynccontextmanager
from itertools import groupby
from typing import *
from types import ModuleType

from pydantic import BaseModel

from .. import KafkaEvent
from .app import FastKafka, AwaitedMock, _get_kafka_brokers
from .._components.asyncapi import KafkaBroker, KafkaBrokers
from .._components.helpers import unwrap_list_type
from .._components.meta import delegates, export, patch
from .._components.producer_decorator import unwrap_from_kafka_event
from .._components.aiokafka_consumer_loop import ConsumeCallable
from .._testing.apache_kafka_broker import ApacheKafkaBroker
from .._testing.in_memory_broker import InMemoryBroker
from .._testing.local_redpanda_broker import LocalRedpandaBroker
from .._components.helpers import remove_suffix

# %% ../../nbs/016_Tester.ipynb 7
def _get_broker_spec(bootstrap_server: str) -> KafkaBroker:
    """
    Helper function to get the broker specification from the bootstrap server URL.

    Args:
        bootstrap_server: The bootstrap server URL in the format "<host>:<port>".

    Returns:
        A KafkaBroker object representing the broker specification.
    """
    url = bootstrap_server.split(":")[0]
    port = bootstrap_server.split(":")[1]
    return KafkaBroker(url=url, port=port, description="", protocol="")

# %% ../../nbs/016_Tester.ipynb 9
@export("fastkafka.testing")
class Tester(FastKafka):
    __test__ = False

    def __init__(
        self,
        app: Union[FastKafka, List[FastKafka]],
        *,
        use_in_memory_broker: bool = True,
    ):
        """Mirror-like object for testing a FastKafka application

        Can be used as context manager

        Args:
            app: The FastKafka application to be tested.
            use_in_memory_broker: Whether to use an in-memory broker for testing or not.
        """
        self.apps = app if isinstance(app, list) else [app]

        for app in self.apps:
            app.create_mocks()

        super().__init__()
        self.mirrors: Dict[Any, Any] = {}
        self._kafka_brokers = self.apps[0]._kafka_brokers
        self._kafka_config["bootstrap_servers_id"] = self.apps[0]._kafka_config[
            "bootstrap_servers_id"
        ]
        self._create_mirrors()
        self.use_in_memory_broker = use_in_memory_broker

    async def _start_tester(self) -> None:
        """Starts the Tester"""
        for app in self.apps:
            await app.__aenter__()
        self.create_mocks()
        self._arrange_mirrors()
        await super().__aenter__()
        await asyncio.sleep(3)

    async def _stop_tester(self) -> None:
        """Shuts down the Tester"""
        await super().__aexit__(None, None, None)
        for app in self.apps[::-1]:
            await app.__aexit__(None, None, None)

    def _create_mirrors(self) -> None:
        pass

    def _arrange_mirrors(self) -> None:
        pass

    @asynccontextmanager
    async def _create_ctx(self) -> AsyncGenerator["Tester", None]:
        if self.use_in_memory_broker == True:
            with InMemoryBroker():  # type: ignore
                await self._start_tester()
                try:
                    yield self
                finally:
                    await self._stop_tester()
        else:
            await self._start_tester()
            try:
                yield self
            finally:
                await self._stop_tester()

    async def __aenter__(self) -> "Tester":
        self._ctx = self._create_ctx()
        return await self._ctx.__aenter__()

    async def __aexit__(self, *args: Any) -> None:
        await self._ctx.__aexit__(*args)

# %% ../../nbs/016_Tester.ipynb 16
def mirror_producer(
    topic: str, producer_f: Callable[..., Any], brokers: str, app: FastKafka
) -> Callable[..., Any]:
    """
    Decorator to create a mirrored producer function.

    Args:
        topic: The topic to produce to.
        producer_f: The original producer function.
        brokers: The brokers configuration.
        app: The FastKafka application.

    Returns:
        The mirrored producer function.
    """
    msg_type = inspect.signature(producer_f).return_annotation

    msg_type_unwrapped = unwrap_list_type(unwrap_from_kafka_event(msg_type))

    async def skeleton_func(msg: BaseModel) -> None:
        pass

    mirror_func = skeleton_func
    sig = inspect.signature(skeleton_func)

    # adjust name, take into consideration the origin app and brokers
    # configuration so that we can differentiate those two
    mirror_func.__name__ = f"mirror_{id(app)}_on_{remove_suffix(topic).replace('.', '_')}_{abs(hash(brokers))}"

    # adjust arg and return val
    sig = sig.replace(
        parameters=[
            inspect.Parameter(
                name="msg",
                annotation=msg_type_unwrapped,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        ]
    )

    mirror_func.__signature__ = sig  # type: ignore

    return mirror_func

# %% ../../nbs/016_Tester.ipynb 19
def mirror_consumer(
    topic: str, consumer_f: Callable[..., Any], brokers: str, app: FastKafka
) -> Callable[[BaseModel], Coroutine[Any, Any, BaseModel]]:
    """
    Decorator to create a mirrored consumer function.

    Args:
        topic: The topic to consume from.
        consumer_f: The original consumer function.
        brokers: The brokers configuration.
        app: The FastKafka application.

    Returns:
        The mirrored consumer function.
    """
    msg_type = inspect.signature(consumer_f).parameters["msg"]

    msg_type_unwrapped = unwrap_list_type(msg_type)

    async def skeleton_func(msg: BaseModel) -> BaseModel:
        return msg

    mirror_func = skeleton_func
    sig = inspect.signature(skeleton_func)

    # adjust name, take into consideration the origin app and brokers
    # configuration so that we can differentiate those two
    mirror_func.__name__ = f"mirror_{id(app)}_to_{remove_suffix(topic).replace('.', '_')}_{abs(hash(brokers))}"

    # adjust arg and return val
    sig = sig.replace(
        parameters=[msg_type], return_annotation=msg_type_unwrapped.annotation
    )

    mirror_func.__signature__ = sig  # type: ignore
    return mirror_func

# %% ../../nbs/016_Tester.ipynb 21
@patch
def _create_mirrors(self: Tester) -> None:
    """
    Creates mirror functions for producers and consumers.

    Iterates over the FastKafka application and its producers and consumers. For each consumer, it creates a mirror
    consumer function using the `mirror_consumer` decorator. For each producer, it creates a mirror producer function
    using the `mirror_producer` decorator. The mirror functions are stored in the `self.mirrors` dictionary and also
    set as attributes on the Tester instance.

    Returns:
        None
    """
    for app in self.apps:
        for topic, (consumer_f, _, _, brokers, _) in app._consumers_store.items():
            mirror_f = mirror_consumer(
                topic,
                consumer_f,
                brokers.model_dump_json()
                if brokers is not None
                else app._kafka_brokers.model_dump_json(),
                app,
            )
            mirror_f = self.produces(  # type: ignore
                topic=remove_suffix(topic),
                brokers=brokers,
            )(mirror_f)
            self.mirrors[consumer_f] = mirror_f
            setattr(self, mirror_f.__name__, mirror_f)
        for topic, (producer_f, _, brokers, _) in app._producers_store.items():
            mirror_f = mirror_producer(
                topic,
                producer_f,
                brokers.model_dump_json()
                if brokers is not None
                else app._kafka_brokers.model_dump_json(),
                app,
            )
            mirror_f = self.consumes(
                topic=remove_suffix(topic),
                brokers=brokers,
            )(
                mirror_f  # type: ignore
            )
            self.mirrors[producer_f] = mirror_f
            setattr(self, mirror_f.__name__, mirror_f)

# %% ../../nbs/016_Tester.ipynb 25
class AmbiguousWarning:
    """
    Warning class used for ambiguous topics.

    Args:
        topic: The ambiguous topic.
        functions: List of function names associated with the ambiguous topic.
    """

    def __init__(self, topic: str, functions: List[str]):
        self.topic = topic
        self.functions = functions

    def __getattribute__(self, attr: str) -> Any:
        raise RuntimeError(
            f"Ambiguous topic: {super().__getattribute__('topic')}, for functions: {super().__getattribute__('functions')}\nUse Tester.mirrors[app.function] to resolve ambiguity"
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise RuntimeError(
            f"Ambiguous topic: {self.topic}, for functions: {self.functions}\nUse Tester.mirrors[app.function] to resolve ambiguity"
        )

# %% ../../nbs/016_Tester.ipynb 27
def set_sugar(
    *,
    tester: Tester,
    prefix: str,
    topic_brokers: Dict[str, Tuple[List[str], List[str]]],
    topic: str,
    brokers: str,
    origin_function_name: str,
    function: Callable[..., Union[Any, Awaitable[Any]]],
) -> None:
    """
    Sets the sugar function for a topic.

    Args:
        tester: The Tester instance.
        prefix: The prefix to use for the sugar function (e.g., "to_" or "on_").
        topic_brokers: Dictionary to store the brokers and functions associated with each topic.
        topic: The topic name.
        brokers: The brokers configuration.
        origin_function_name: The name of the original function.
        function: The mirror function to be set as the sugar function.

    Returns:
        None
    """
    brokers_for_topic, functions_for_topic = topic_brokers.get(topic, ([], []))
    if brokers not in brokers_for_topic:
        brokers_for_topic.append(brokers)
        functions_for_topic.append(origin_function_name)
        topic_brokers[topic] = (brokers_for_topic, functions_for_topic)
    if len(brokers_for_topic) == 1:
        setattr(tester, f"{prefix}{topic}", function)
    else:
        setattr(
            tester, f"{prefix}{topic}", AmbiguousWarning(topic, functions_for_topic)
        )

# %% ../../nbs/016_Tester.ipynb 28
@patch
def _arrange_mirrors(self: Tester) -> None:
    """
    Arranges the mirror functions.

    Iterates over the FastKafka application and its producers and consumers. For each consumer, it retrieves the mirror
    function from the `self.mirrors` dictionary and sets it as an attribute on the Tester instance. It also sets the
    sugar function using the `set_sugar` function. For each producer, it retrieves the mirror function and sets it as
    an attribute on the Tester instance. It also sets the sugar function for the awaited mocks. Finally, it creates the
    `mocks` and `awaited_mocks` namedtuples and sets them as attributes on the Tester instance.

    Returns:
        None
    """
    topic_brokers: Dict[str, Tuple[List[str], List[str]]] = {}
    mocks = {}
    awaited_mocks = {}
    for app in self.apps:
        for topic, (consumer_f, _, _, brokers, _) in app._consumers_store.items():
            mirror_f = self.mirrors[consumer_f]
            self.mirrors[getattr(app, consumer_f.__name__)] = mirror_f
            set_sugar(
                tester=self,
                prefix="to_",
                topic_brokers=topic_brokers,
                topic=remove_suffix(topic),
                brokers=brokers.model_dump_json()
                if brokers is not None
                else app._kafka_brokers.model_dump_json(),
                origin_function_name=consumer_f.__name__,
                function=mirror_f,
            )

            mocks[f"to_{remove_suffix(topic)}"] = getattr(self.mocks, mirror_f.__name__)
            awaited_mocks[f"to_{remove_suffix(topic)}"] = getattr(
                self.awaited_mocks, mirror_f.__name__
            )

        for topic, (producer_f, _, brokers, _) in app._producers_store.items():
            mirror_f = self.mirrors[producer_f]
            self.mirrors[getattr(app, producer_f.__name__)] = getattr(
                self.awaited_mocks, mirror_f.__name__
            )
            set_sugar(
                tester=self,
                prefix="on_",
                topic_brokers=topic_brokers,
                topic=remove_suffix(topic),
                brokers=brokers.model_dump_json()
                if brokers is not None
                else app._kafka_brokers.model_dump_json(),
                origin_function_name=producer_f.__name__,
                function=getattr(self.awaited_mocks, mirror_f.__name__),
            )
            mocks[f"on_{remove_suffix(topic)}"] = getattr(self.mocks, mirror_f.__name__)
            awaited_mocks[f"on_{remove_suffix(topic)}"] = getattr(
                self.awaited_mocks, mirror_f.__name__
            )

    AppMocks = collections.namedtuple(  # type: ignore
        f"{self.__class__.__name__}Mocks", [f_name for f_name in mocks]
    )
    setattr(self, "mocks", AppMocks(**mocks))
    setattr(self, "awaited_mocks", AppMocks(**awaited_mocks))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/011_ConsumerLoop.ipynb.

# %% auto 0
__all__ = ['logger', 'AsyncConsume', 'AsyncConsumeMeta', 'SyncConsume', 'SyncConsumeMeta', 'ConsumeCallable', 'EventMetadata',
           'sanitize_kafka_config', 'aiokafka_consumer_loop']

# %% ../../nbs/011_ConsumerLoop.ipynb 1
from asyncio import iscoroutinefunction, Task  # do not use the version from inspect
from typing import *
from dataclasses import dataclass

import asyncer
from aiokafka.structs import ConsumerRecord
from pydantic import BaseModel

import fastkafka._aiokafka_imports
from .logger import get_logger
from .meta import delegates, export
from .task_streaming import get_executor, StreamExecutor

# %% ../../nbs/011_ConsumerLoop.ipynb 5
logger = get_logger(__name__)

# %% ../../nbs/011_ConsumerLoop.ipynb 8
@dataclass
@export("fastkafka")
class EventMetadata:
    """A class for encapsulating Kafka record metadata.

    Args:
        topic: The topic this record is received from
        partition: The partition from which this record is received
        offset: The position of this record in the corresponding Kafka partition
        timestamp: The timestamp of this record
        timestamp_type: The timestamp type of this record
        key: The key (or `None` if no key is specified)
        value: The value
        serialized_key_size: The size of the serialized, uncompressed key in bytes
        serialized_value_size: The size of the serialized, uncompressed value in bytes
        headers: The headers
    """

    topic: str
    partition: int
    offset: int
    timestamp: int
    timestamp_type: int
    key: Optional[bytes]
    value: Optional[bytes]
    checksum: int
    serialized_key_size: int
    serialized_value_size: int
    headers: Sequence[Tuple[str, bytes]]

    @staticmethod
    def create_event_metadata(record: ConsumerRecord) -> "EventMetadata":  # type: ignore
        """Creates an instance of EventMetadata from a ConsumerRecord.

        Args:
            record: The Kafka ConsumerRecord.

        Returns:
            The created EventMetadata instance.
        """
        return EventMetadata(
            topic=record.topic,
            partition=record.partition,
            offset=record.offset,
            timestamp=record.timestamp,
            timestamp_type=record.timestamp_type,
            value=record.value,
            checksum=record.checksum,
            key=record.key,
            serialized_key_size=record.serialized_key_size,
            serialized_value_size=record.serialized_value_size,
            headers=record.headers,
        )

# %% ../../nbs/011_ConsumerLoop.ipynb 11
AsyncConsume = Callable[[Union[List[BaseModel], BaseModel]], Awaitable[None]]
AsyncConsumeMeta = Callable[
    [Union[List[BaseModel], BaseModel], Union[List[EventMetadata], EventMetadata]],
    Awaitable[None],
]
SyncConsume = Callable[[Union[List[BaseModel], BaseModel]], None]
SyncConsumeMeta = Callable[
    [Union[List[BaseModel], BaseModel], Union[List[EventMetadata], EventMetadata]], None
]

ConsumeCallable = Union[AsyncConsume, AsyncConsumeMeta, SyncConsume, SyncConsumeMeta]

# %% ../../nbs/011_ConsumerLoop.ipynb 12
def _callback_parameters_wrapper(
    callback: Union[AsyncConsume, AsyncConsumeMeta]
) -> AsyncConsumeMeta:
    """Wraps an async callback and filters the arguments to pass based on if the function accepts EventMetadata as argument

    Args:
        callback: async callable that will be wrapped

    Returns:
        Wrapped callback with filtered params
    """

    async def _params_wrap(
        msg: Union[BaseModel, List[BaseModel]],
        meta: Union[EventMetadata, List[EventMetadata]],
        callback: Union[AsyncConsume, AsyncConsumeMeta] = callback,
    ) -> None:
        types = list(get_type_hints(callback).values())
        args: List[
            Union[BaseModel, List[BaseModel], EventMetadata, List[EventMetadata]]
        ] = [msg]
        if EventMetadata in types:
            args.insert(types.index(EventMetadata), meta)
        if List[EventMetadata] in types:
            args.insert(types.index(List[EventMetadata]), meta)
        await callback(*args)  # type: ignore

    return _params_wrap

# %% ../../nbs/011_ConsumerLoop.ipynb 17
def _prepare_callback(callback: ConsumeCallable) -> AsyncConsumeMeta:
    """
    Prepares a callback to be used in the consumer loop.
        1. If callback is sync, asyncify it
        2. Wrap the callback into a safe callback for exception handling

    Args:
        callback: async callable that will be prepared for use in consumer

    Returns:
        Prepared callback
    """
    async_callback: Union[AsyncConsume, AsyncConsumeMeta] = (
        callback if iscoroutinefunction(callback) else asyncer.asyncify(callback)  # type: ignore
    )
    return _callback_parameters_wrapper(async_callback)

# %% ../../nbs/011_ConsumerLoop.ipynb 24
def _get_single_msg_handlers(  # type: ignore
    *,
    consumer: fastkafka._aiokafka_imports.AIOKafkaConsumer,
    callback: AsyncConsumeMeta,
    decoder_fn: Callable[[bytes, Type[BaseModel]], Any],
    msg_type: Type[BaseModel],
    **kwargs: Any,
) -> Tuple[
    Callable[
        [
            ConsumerRecord,
            AsyncConsumeMeta,
            Callable[[bytes, Type[BaseModel]], Any],
            Type[BaseModel],
        ],
        Awaitable[None],
    ],
    Callable[
        [fastkafka._aiokafka_imports.AIOKafkaConsumer, Any],
        Awaitable[List[ConsumerRecord]],
    ],
]:
    """
    Retrieves the message handlers for consuming single messages from a Kafka topic.

    Args:
        consumer: The Kafka consumer instance.
        callback: The callback function to handle the consumed message.
        decoder_fn: The function to decode the consumed message.
        msg_type: The type of the consumed message.
        **kwargs: Additional keyword arguments for the consumer.

    Returns:
        The handle_msg function and poll_consumer function.
    """

    async def handle_msg(  # type: ignore
        record: ConsumerRecord,
        callback: AsyncConsumeMeta = callback,
        decoder_fn: Callable[[bytes, Type[BaseModel]], Any] = decoder_fn,
        msg_type: Type[BaseModel] = msg_type,
    ) -> None:
        await callback(
            decoder_fn(record.value, msg_type),
            EventMetadata.create_event_metadata(record),
        )

    async def poll_consumer(  # type: ignore
        consumer: fastkafka._aiokafka_imports.AIOKafkaConsumer = consumer,
        kwargs: Any = kwargs,
    ) -> List[ConsumerRecord]:
        msgs = await consumer.getmany(**kwargs)
        return [msg for msg_group in msgs.values() for msg in msg_group]

    return handle_msg, poll_consumer

# %% ../../nbs/011_ConsumerLoop.ipynb 26
def _get_batch_msg_handlers(  # type: ignore
    *,
    consumer: fastkafka._aiokafka_imports.AIOKafkaConsumer,
    callback: AsyncConsumeMeta,
    decoder_fn: Callable[[bytes, Type[BaseModel]], Any],
    msg_type: Type[BaseModel],
    **kwargs: Any,
) -> Tuple[
    Callable[
        [
            List[ConsumerRecord],
            AsyncConsumeMeta,
            Callable[[bytes, Type[BaseModel]], Any],
            Type[BaseModel],
        ],
        Awaitable[None],
    ],
    Callable[
        [fastkafka._aiokafka_imports.AIOKafkaConsumer, Any],
        Awaitable[List[List[ConsumerRecord]]],
    ],
]:
    """
    Retrieves the message handlers for consuming messages in batches from a Kafka topic.

    Args:
        consumer: The Kafka consumer instance.
        callback: The callback function to handle the consumed messages.
        decoder_fn: The function to decode the consumed messages.
        msg_type: The type of the consumed messages.
        **kwargs: Additional keyword arguments for the consumer.

    Returns:
        The handle_msg function and poll_consumer function.
    """

    async def handle_msg(  # type: ignore
        records: List[ConsumerRecord],
        callback: AsyncConsumeMeta = callback,
        decoder_fn: Callable[[bytes, Type[BaseModel]], Any] = decoder_fn,
        msg_type: Type[BaseModel] = msg_type,
    ) -> None:
        await callback(
            [decoder_fn(record.value, msg_type) for record in records],
            [EventMetadata.create_event_metadata(record) for record in records],
        )

    async def poll_consumer(  # type: ignore
        consumer: fastkafka._aiokafka_imports.AIOKafkaConsumer = consumer,
        kwargs: Any = kwargs,
    ) -> List[List[ConsumerRecord]]:
        msgs = await consumer.getmany(**kwargs)
        return [value for value in msgs.values() if len(value) > 0]

    return handle_msg, poll_consumer

# %% ../../nbs/011_ConsumerLoop.ipynb 28
@delegates(fastkafka._aiokafka_imports.AIOKafkaConsumer.getmany)
async def _aiokafka_consumer_loop(  # type: ignore
    consumer: fastkafka._aiokafka_imports.AIOKafkaConsumer,
    *,
    topic: str,
    decoder_fn: Callable[[bytes, Type[BaseModel]], Any],
    callback: ConsumeCallable,
    max_buffer_size: int = 100_000,
    msg_type: Union[Type[List[BaseModel]], Type[BaseModel]],
    is_shutting_down_f: Callable[[], bool],
    executor: Union[str, StreamExecutor, None] = None,
    **kwargs: Any,
) -> None:
    """
    Consumer loop for infinite pooling of the AIOKafka consumer for new messages. Calls consumer.getmany()
    and after the consumer return messages or times out, messages are decoded and streamed to defined callback.

    Args:
        topic: Topic to subscribe
        decoder_fn: Function to decode the messages consumed from the topic
        callbacks: Dict of callbacks mapped to their respective topics
        timeout_ms: Time to timeut the getmany request by the consumer
        max_buffer_size: Maximum number of unconsumed messages in the callback buffer
        msg_types: Dict of message types mapped to their respective topics
        is_shutting_down_f: Function for controlling the shutdown of consumer loop
    """

    prepared_callback = _prepare_callback(callback)

    if hasattr(msg_type, "__origin__") and msg_type.__origin__ == list:
        handle_msg, poll_consumer = _get_batch_msg_handlers(
            consumer=consumer,
            callback=prepared_callback,
            decoder_fn=decoder_fn,
            msg_type=msg_type.__args__[0],  # type: ignore
            **kwargs,
        )
    else:
        handle_msg, poll_consumer = _get_single_msg_handlers(
            consumer=consumer,
            callback=prepared_callback,
            decoder_fn=decoder_fn,
            msg_type=msg_type,  # type: ignore
            **kwargs,
        )

    await get_executor(executor).run(
        is_shutting_down_f=is_shutting_down_f,
        generator=poll_consumer,  # type: ignore
        processor=handle_msg,  # type: ignore
    )

# %% ../../nbs/011_ConsumerLoop.ipynb 35
def sanitize_kafka_config(**kwargs: Any) -> Dict[str, Any]:
    """Sanitize Kafka config"""
    return {k: "*" * len(v) if "pass" in k.lower() else v for k, v in kwargs.items()}

# %% ../../nbs/011_ConsumerLoop.ipynb 37
@delegates(fastkafka._aiokafka_imports.AIOKafkaConsumer)
@delegates(_aiokafka_consumer_loop, keep=True)
async def aiokafka_consumer_loop(
    topic: str,
    decoder_fn: Callable[[bytes, Type[BaseModel]], Any],
    *,
    timeout_ms: int = 100,
    max_buffer_size: int = 100_000,
    callback: ConsumeCallable,
    msg_type: Union[Type[List[BaseModel]], Type[BaseModel]],
    is_shutting_down_f: Callable[[], bool],
    executor: Union[str, StreamExecutor, None] = None,
    **kwargs: Any,
) -> None:
    """Consumer loop for infinite pooling of the AIOKafka consumer for new messages. Creates and starts AIOKafkaConsumer
    and runs _aio_kafka_consumer loop fo infinite poling of the consumer for new messages.

    Args:
        topic: name of the topic to subscribe to
        decoder_fn: Function to decode the messages consumed from the topic
        callback: callback function to be called after decoding and parsing a consumed message
        timeout_ms: Time to timeut the getmany request by the consumer
        max_buffer_size: Maximum number of unconsumed messages in the callback buffer
        msg_type: Type with `parse_json` method used for parsing a decoded message
        is_shutting_down_f: Function for controlling the shutdown of consumer loop
    """
    logger.info(f"aiokafka_consumer_loop() starting...")
    try:
        consumer = fastkafka._aiokafka_imports.AIOKafkaConsumer(
            **kwargs,
        )
        logger.info(
            f"aiokafka_consumer_loop(): Consumer created using the following parameters: {sanitize_kafka_config(**kwargs)}"
        )

        await consumer.start()
        logger.info("aiokafka_consumer_loop(): Consumer started.")
        consumer.subscribe([topic])
        logger.info("aiokafka_consumer_loop(): Consumer subscribed.")

        try:
            await _aiokafka_consumer_loop(
                consumer=consumer,
                topic=topic,
                decoder_fn=decoder_fn,
                max_buffer_size=max_buffer_size,
                timeout_ms=timeout_ms,
                callback=callback,
                msg_type=msg_type,
                is_shutting_down_f=is_shutting_down_f,
                executor=executor,
            )
        finally:
            await consumer.stop()
            logger.info(f"aiokafka_consumer_loop(): Consumer stopped.")
            logger.info(f"aiokafka_consumer_loop() finished.")
    except Exception as e:
        logger.error(
            f"aiokafka_consumer_loop(): unexpected exception raised: '{e.__repr__()}'"
        )
        raise e



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/006_TaskStreaming.ipynb.

# %% auto 0
__all__ = ['logger', 'TaskPool', 'ExceptionMonitor', 'StreamExecutor', 'DynamicTaskExecutor', 'SequentialExecutor',
           'get_executor']

# %% ../../nbs/006_TaskStreaming.ipynb 1
import asyncio
import sys
from abc import ABC, abstractmethod

from asyncio import Task
from typing import *

import anyio
from aiokafka import ConsumerRecord

from logging import Logger
from .logger import get_logger
from .meta import export

# %% ../../nbs/006_TaskStreaming.ipynb 3
logger = get_logger(__name__)

# %% ../../nbs/006_TaskStreaming.ipynb 10
class TaskPool:
    def __init__(
        self,
        size: int = 100_000,
        on_error: Optional[Callable[[BaseException], None]] = None,
    ):
        """
        Initializes a TaskPool instance.

        Args:
            size: The size of the task pool. Defaults to 100,000.
            on_error: Optional callback function to handle task errors. Defaults to None.

        Returns:
            None
        """
        self.size = size
        self.pool: Set[Task] = set()
        self.on_error = on_error
        self.finished = False

    async def add(self, item: Task) -> None:
        """
        Adds a task to the task pool.

        Args:
            item: The task to be added.

        Returns:
            None
        """
        while len(self.pool) >= self.size:
            await asyncio.sleep(0)
        self.pool.add(item)
        item.add_done_callback(self.discard)

    def discard(self, task: Task) -> None:
        """
        Discards a completed task from the task pool.

        Args:
            task: The completed task to be discarded.

        Returns:
            None
        """
        e = task.exception()
        if e is not None and self.on_error is not None:
            try:
                self.on_error(e)
            except Exception as ee:
                logger.warning(
                    f"Exception {ee} raised when calling on_error() callback: {e}"
                )

        self.pool.discard(task)

    def __len__(self) -> int:
        """
        Returns the number of tasks in the task pool.

        Returns:
            The number of tasks in the task pool.
        """
        return len(self.pool)

    async def __aenter__(self) -> "TaskPool":
        self.finished = False
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        while len(self) > 0:
            await asyncio.sleep(0)
        self.finished = True

    @staticmethod
    def log_error(logger: Logger) -> Callable[[Exception], None]:
        """
        Creates a decorator that logs errors using the specified logger.

        Args:
            logger: The logger to use for error logging.

        Returns:
            The decorator function.
        """

        def _log_error(e: Exception, logger: Logger = logger) -> None:
            logger.warning(f"{e=}")

        return _log_error

# %% ../../nbs/006_TaskStreaming.ipynb 14
class ExceptionMonitor:
    def __init__(self) -> None:
        """
        Initializes an ExceptionMonitor instance.

        Returns:
            None
        """
        self.exceptions: List[Exception] = []
        self.exception_found = False

    def on_error(self, e: Exception) -> None:
        """
        Handles an error by storing the exception.

        Args:
            e: The exception to be handled.

        Returns:
            None
        """
        self.exceptions.append(e)
        self.exception_found = True

    def _monitor_step(self) -> None:
        """
        Raises the next exception in the queue.

        Returns:
            None
        """
        if len(self.exceptions) > 0:
            e = self.exceptions.pop(0)
            raise e

    async def __aenter__(self) -> "ExceptionMonitor":
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        while len(self.exceptions) > 0:
            self._monitor_step()
            await asyncio.sleep(0)

# %% ../../nbs/006_TaskStreaming.ipynb 17
class StreamExecutor(ABC):
    @abstractmethod
    async def run(  # type: ignore
        self,
        *,
        is_shutting_down_f: Callable[[], bool],
        generator: Callable[[], Awaitable[ConsumerRecord]],
        processor: Callable[[ConsumerRecord], Awaitable[None]],
    ) -> None:
        """
        Abstract method for running the stream executor.

        Args:
            is_shutting_down_f: Function to check if the executor is shutting down.
            generator: Generator function for retrieving consumer records.
            processor: Processor function for processing consumer records.
        """
        pass

# %% ../../nbs/006_TaskStreaming.ipynb 20
def _process_items_task(  # type: ignore
    processor: Callable[[ConsumerRecord], Awaitable[None]], task_pool: TaskPool
) -> Callable[
    [
        anyio.streams.memory.MemoryObjectReceiveStream,
        Callable[[ConsumerRecord], Awaitable[None]],
        bool,
    ],
    Coroutine[Any, Any, Awaitable[None]],
]:
    async def _process_items_wrapper(  # type: ignore
        receive_stream: anyio.streams.memory.MemoryObjectReceiveStream,
        processor: Callable[[ConsumerRecord], Awaitable[None]] = processor,
        task_pool=task_pool,
    ):
        async with receive_stream:
            async for msg in receive_stream:
                task: asyncio.Task = asyncio.create_task(processor(msg))  # type: ignore
                await task_pool.add(task)

    return _process_items_wrapper

# %% ../../nbs/006_TaskStreaming.ipynb 21
@export("fastkafka.executors")
class DynamicTaskExecutor(StreamExecutor):
    """A class that implements a dynamic task executor for processing consumer records.

    The DynamicTaskExecutor class extends the StreamExecutor class and provides functionality
    for running a tasks in parallel using asyncio.Task.
    """

    def __init__(
        self,
        throw_exceptions: bool = False,
        max_buffer_size: int = 100_000,
        size: int = 100_000,
    ):
        """Create an instance of DynamicTaskExecutor

        Args:
            throw_exceptions: Flag indicating whether exceptions should be thrown ot logged.
                Defaults to False.
            max_buffer_size: Maximum buffer size for the memory object stream.
                Defaults to 100_000.
            size: Size of the task pool. Defaults to 100_000.
        """
        self.throw_exceptions = throw_exceptions
        self.max_buffer_size = max_buffer_size
        self.exception_monitor = ExceptionMonitor()
        self.task_pool = TaskPool(
            on_error=self.exception_monitor.on_error  # type: ignore
            if throw_exceptions
            else TaskPool.log_error(logger),
            size=size,
        )

    async def run(  # type: ignore
        self,
        *,
        is_shutting_down_f: Callable[[], bool],
        generator: Callable[[], Awaitable[ConsumerRecord]],
        processor: Callable[[ConsumerRecord], Awaitable[None]],
    ) -> None:
        """
        Runs the dynamic task executor.

        Args:
            is_shutting_down_f: Function to check if the executor is shutting down.
            generator: Generator function for retrieving consumer records.
            processor: Processor function for processing consumer records.
        """
        send_stream, receive_stream = anyio.create_memory_object_stream(
            max_buffer_size=self.max_buffer_size
        )

        async with self.exception_monitor, self.task_pool:
            async with anyio.create_task_group() as tg:
                tg.start_soon(
                    _process_items_task(processor, self.task_pool), receive_stream
                )
                async with send_stream:
                    while not is_shutting_down_f():
                        if (
                            self.exception_monitor.exception_found
                            and self.throw_exceptions
                        ):
                            break
                        msgs = await generator()
                        for msg in msgs:
                            await send_stream.send(msg)

# %% ../../nbs/006_TaskStreaming.ipynb 30
def _process_items_coro(  # type: ignore
    processor: Callable[[ConsumerRecord], Awaitable[None]],
    throw_exceptions: bool,
) -> Callable[
    [
        anyio.streams.memory.MemoryObjectReceiveStream,
        Callable[[ConsumerRecord], Awaitable[None]],
        bool,
    ],
    Coroutine[Any, Any, Awaitable[None]],
]:
    async def _process_items_wrapper(  # type: ignore
        receive_stream: anyio.streams.memory.MemoryObjectReceiveStream,
        processor: Callable[[ConsumerRecord], Awaitable[None]] = processor,
        throw_exceptions: bool = throw_exceptions,
    ) -> Awaitable[None]:
        async with receive_stream:
            async for msg in receive_stream:
                try:
                    await processor(msg)
                except Exception as e:
                    if throw_exceptions:
                        raise e
                    else:
                        logger.warning(f"{e=}")

    return _process_items_wrapper

# %% ../../nbs/006_TaskStreaming.ipynb 31
@export("fastkafka.executors")
class SequentialExecutor(StreamExecutor):
    """A class that implements a sequential executor for processing consumer records.

    The SequentialExecutor class extends the StreamExecutor class and provides functionality
    for running processing tasks in sequence by awaiting their coroutines.
    """

    def __init__(
        self,
        throw_exceptions: bool = False,
        max_buffer_size: int = 100_000,
    ):
        """Create an instance of SequentialExecutor

        Args:
            throw_exceptions: Flag indicating whether exceptions should be thrown or logged.
                Defaults to False.
            max_buffer_size: Maximum buffer size for the memory object stream.
                Defaults to 100_000.
        """
        self.throw_exceptions = throw_exceptions
        self.max_buffer_size = max_buffer_size

    async def run(  # type: ignore
        self,
        *,
        is_shutting_down_f: Callable[[], bool],
        generator: Callable[[], Awaitable[ConsumerRecord]],
        processor: Callable[[ConsumerRecord], Awaitable[None]],
    ) -> None:
        """
        Runs the sequential executor.

        Args:
            is_shutting_down_f: Function to check if the executor is shutting down.
            generator: Generator function for retrieving consumer records.
            processor: Processor function for processing consumer records.
        """

        send_stream, receive_stream = anyio.create_memory_object_stream(
            max_buffer_size=self.max_buffer_size
        )

        async with anyio.create_task_group() as tg:
            tg.start_soon(
                _process_items_coro(processor, self.throw_exceptions), receive_stream
            )
            async with send_stream:
                while not is_shutting_down_f():
                    msgs = await generator()
                    for msg in msgs:
                        await send_stream.send(msg)

# %% ../../nbs/006_TaskStreaming.ipynb 34
def get_executor(executor: Union[str, StreamExecutor, None] = None) -> StreamExecutor:
    """
    Returns an instance of the specified executor.

    Args:
        executor: Executor instance or name of the executor.

    Returns:
        Instance of the specified executor.

    Raises:
        AttributeError: If the executor is not found.
    """
    if isinstance(executor, StreamExecutor):
        return executor
    elif executor is None:
        executor = "SequentialExecutor"
    return getattr(sys.modules["fastkafka._components.task_streaming"], executor)()  # type: ignore



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/098_Test_Dependencies.ipynb.

# %% auto 0
__all__ = ['logger', 'kafka_repo_url', 'local_path', 'check_java', 'VersionParser', 'get_kafka_version', 'check_kafka',
           'generate_app_src', 'generate_app_in_tmp']

# %% ../../nbs/098_Test_Dependencies.ipynb 2
import re
import platform
import shutil
import tarfile
from contextlib import contextmanager
from html.parser import HTMLParser
from os import environ, rename
from os.path import expanduser
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import *

from packaging import version

from .helpers import change_dir, in_notebook
from .logger import get_logger

if in_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

# %% ../../nbs/098_Test_Dependencies.ipynb 4
logger = get_logger(__name__)

# %% ../../nbs/098_Test_Dependencies.ipynb 6
def check_java(*, potential_jdk_path: Optional[List[Path]] = None) -> bool:
    """Checks if JDK 11 is installed on the machine and exports it to PATH if necessary.

    Args:
        potential_jdk_path: Optional. List of potential paths where JDK 11 may be installed.
                            If not provided, it defaults to searching for JDK 11 in the user's home directory.

    Returns:
        bool: True if JDK 11 is installed and exported to PATH, False otherwise.
    """
    if potential_jdk_path is None:
        potential_jdk_path = list(Path(expanduser("~") + "/.jdk").glob("jdk-11*"))

    if potential_jdk_path != []:
        logger.info("Java is already installed.")
        if not shutil.which("java"):
            logger.info("But not exported to PATH, exporting...")
            env_path_separator = ";" if platform.system() == "Windows" else ":"
            environ["PATH"] = (
                environ["PATH"] + f"{env_path_separator}{potential_jdk_path[0]/ 'bin'}"
            )
        return True
    return False

# %% ../../nbs/098_Test_Dependencies.ipynb 8
def _install_java() -> None:
    """Checks if jdk-11 is installed on the machine and installs it if not

    Returns:
       None

    Raises:
        RuntimeError: If JDK 11 installation fails.
    """
    try:
        import jdk
    except Exception as e:
        msg = "Please install test version of fastkafka using 'pip install fastkafka[test]' command"
        logger.error(msg)
        raise RuntimeError(msg)

    if not check_java():
        logger.info("Installing Java...")
        logger.info(" - installing jdk...")
        jdk_bin_path = Path(jdk.install("11"))
        logger.info(f" - jdk path: {jdk_bin_path}")
        env_path_separator = ";" if platform.system() == "Windows" else ":"
        environ["PATH"] = (
            environ["PATH"] + f"{env_path_separator}{jdk_bin_path / 'bin'}"
        )
        logger.info("Java installed.")

# %% ../../nbs/098_Test_Dependencies.ipynb 10
class VersionParser(HTMLParser):
    """
    A parser class for extracting the newest version of a software from HTML data.

    This class extends the HTMLParser class and provides a mechanism to extract the newest version of a software
    from HTML data using regular expressions.

    Attributes:
        newest_version (str): The newest version of the software.

    """

    def __init__(self) -> None:
        """
        Initializes a VersionParser object.

        The newest_version attribute is initialized to "0.0.0".

        """
        HTMLParser.__init__(self)
        self.newest_version = "0.0.0"

    def handle_data(self, data: str) -> None:
        """
        Handles the data encountered in the HTML parsing process.

        This method is called by the HTMLParser base class when data is encountered within HTML tags.
        It uses regular expressions to search for version numbers in the data and updates the newest_version
        attribute if a higher version is found.

        Args:
            data (str): The data encountered during parsing.

        """
        match = re.search("[0-9]+\.[0-9]+\.[0-9]+", data)
        if match is not None:
            if version.parse(self.newest_version) < version.parse(match.group(0)):
                self.newest_version = match.group(0)

# %% ../../nbs/098_Test_Dependencies.ipynb 12
kafka_repo_url = "https://dlcdn.apache.org/kafka"


def get_kafka_version(kafka_repo_url: str = kafka_repo_url) -> str:
    """
    Retrieves the newest version of Kafka from the given Kafka repository URL.

    Args:
        kafka_repo_url: The URL of the Kafka repository. Defaults to `https://dlcdn.apache.org/kafka`.

    Returns:
        The newest version of Kafka as a string.

    Raises:
        RuntimeError: If the requests module is not installed or encounters an error during the request.

    """
    try:
        import requests
    except Exception as e:
        msg = "Please install test version of fastkafka using 'pip install fastkafka[test]' command"
        logger.error(msg)
        raise RuntimeError(msg)

    parser = VersionParser()

    response = requests.get(
        kafka_repo_url,
        timeout=60,
    )
    parser.feed(response.text)

    return parser.newest_version

# %% ../../nbs/098_Test_Dependencies.ipynb 13
local_path = (
    Path(expanduser("~")).parent / "Public"
    if platform.system() == "Windows"
    else Path(expanduser("~")) / ".local"
)

# %% ../../nbs/098_Test_Dependencies.ipynb 14
def check_kafka(local_path: Path = local_path) -> bool:
    """Checks if Kafka is installed on the machine and exports it to PATH if necessary.

    Args:
        kafka_path: Path to the Kafka installation directory. Defaults to the global variable `kafka_path`.

    Returns:
        bool: True if Kafka is installed and exported to PATH, False otherwise.
    """

    kafka_fname = f"kafka_2.13-{get_kafka_version()}"

    kafka_path = (
        local_path / "kafka"
        if platform.system() == "Windows"
        else local_path / f"{kafka_fname}"
    )

    if (kafka_path / "bin").exists():
        logger.info("Kafka is installed.")
        if not shutil.which("kafka-server-start.sh"):
            logger.info("But not exported to PATH, exporting...")
            kafka_binary_path = (
                f";{kafka_path / 'bin' / 'windows'}"
                if platform.system() == "Windows"
                else f":{kafka_path / 'bin'}"
            )
            environ["PATH"] = environ["PATH"] + kafka_binary_path
        return True
    return False

# %% ../../nbs/098_Test_Dependencies.ipynb 15
def _install_kafka(
    local_path: Path = local_path, kafka_repo_url: str = kafka_repo_url
) -> None:
    """Checks if Kafka is installed on the machine and installs it if not.

    Args:
        local_path: Path where the Kafka installation package will be stored. Defaults to the global variable `local_path`.
        kafka_repo_url: The URL of the Kafka repository. Defaults to `https://dlcdn.apache.org/kafka`.
    Returns:
       None

    Raises:
        RuntimeError: If Kafka installation fails.
    """
    try:
        import requests
    except Exception as e:
        msg = "Please install test version of fastkafka using 'pip install fastkafka[test]' command"
        logger.error(msg)
        raise RuntimeError(msg)

    kafka_version = get_kafka_version()
    kafka_fname = f"kafka_2.13-{kafka_version}"
    kafka_url = f"{kafka_repo_url}/{kafka_version}/{kafka_fname}.tgz"
    tgz_path = local_path / f"{kafka_fname}.tgz"
    kafka_path = (
        local_path / "kafka"
        if platform.system() == "Windows"
        else local_path / f"{kafka_fname}"
    )

    if not check_kafka():
        logger.info("Installing Kafka...")
        local_path.mkdir(exist_ok=True, parents=True)
        response = requests.get(
            kafka_url,
            stream=True,
            timeout=60,
        )
        try:
            total = response.raw.length_remaining // 128
        except Exception:
            total = None

        with open(tgz_path, "wb") as f:
            for data in tqdm(response.iter_content(chunk_size=128), total=total):
                f.write(data)

        with tarfile.open(tgz_path) as tar:
            for tarinfo in tar:
                tar.extract(tarinfo, local_path)

        if platform.system() == "Windows":
            rename(local_path / f"{kafka_fname}", kafka_path)

        kafka_binary_path = (
            f";{kafka_path / 'bin' / 'windows'}"
            if platform.system() == "Windows"
            else f":{kafka_path / 'bin'}"
        )
        environ["PATH"] = environ["PATH"] + kafka_binary_path
        logger.info(f"Kafka installed in {kafka_path}.")

# %% ../../nbs/098_Test_Dependencies.ipynb 17
def _install_testing_deps() -> None:
    """Installs Java and Kafka dependencies required for testing.

    Raises:
        RuntimeError: If Java or Kafka installation fails.
    """
    _install_java()
    _install_kafka()

# %% ../../nbs/098_Test_Dependencies.ipynb 19
def generate_app_src(out_path: Union[Path, str]) -> None:
    """Generates the source code for the test application based on a Jupyter notebook.

    Args:
        out_path: Path where the generated source code will be saved.

    Raises:
        ValueError: If the Jupyter notebook file does not exist.
    """
    import nbformat
    from nbconvert import PythonExporter

    path = Path("099_Test_Service.ipynb")
    if not path.exists():
        path = Path("..") / "099_Test_Service.ipynb"
    if not path.exists():
        raise ValueError(f"Path '{path.resolve()}' does not exists.")

    with open(path, "r") as f:
        notebook = nbformat.reads(f.read(), nbformat.NO_CONVERT)
        exporter = PythonExporter()
        source, _ = exporter.from_notebook_node(notebook)

    with open(out_path, "w") as f:
        f.write(source)

# %% ../../nbs/098_Test_Dependencies.ipynb 21
@contextmanager
def generate_app_in_tmp() -> Generator[str, None, None]:
    """Context manager that generates the test application source code in a temporary directory.

    Yields:
        str: Import statement for the generated test application.
    """
    with TemporaryDirectory() as d:
        src_path = Path(d) / "main.py"
        generate_app_src(src_path)
        with change_dir(d):
            import_str = f"{src_path.stem}:kafka_app"
            yield import_str



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/018_Avro_Encode_Decoder.ipynb.

# %% auto 0
__all__ = ['logger', 'AvroBase', 'avro_encoder', 'avro_decoder', 'avsc_to_pydantic']

# %% ../../../nbs/018_Avro_Encode_Decoder.ipynb 1
import io
import json
from typing import *

import fastavro
from pydantic import BaseModel, create_model

from ..logger import get_logger
from ..meta import export

# %% ../../../nbs/018_Avro_Encode_Decoder.ipynb 4
logger = get_logger(__name__)

# %% ../../../nbs/018_Avro_Encode_Decoder.ipynb 7
@export("fastkafka.encoder")
class AvroBase(BaseModel):
    """This is base pydantic class that will add some methods"""

    @classmethod
    def avro_schema_for_pydantic_object(
        cls,
        pydantic_model: BaseModel,
        by_alias: bool = True,
        namespace: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Returns the Avro schema for the given Pydantic object.

        Args:
            pydantic_model (BaseModel): The Pydantic object.
            by_alias (bool, optional): Generate schemas using aliases defined. Defaults to True.
            namespace (Optional[str], optional): Optional namespace string for schema generation.

        Returns:
            Dict[str, Any]: The Avro schema for the model.
        """

        schema = pydantic_model.__class__.model_json_schema(by_alias=by_alias)

        if namespace is None:
            # default namespace will be based on title
            namespace = schema["title"]

        return cls._avro_schema(schema, namespace)

    @classmethod
    def avro_schema_for_pydantic_class(
        cls,
        pydantic_model: Type[BaseModel],
        by_alias: bool = True,
        namespace: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Returns the Avro schema for the given Pydantic class.

        Args:
            pydantic_model (Type[BaseModel]): The Pydantic class.
            by_alias (bool, optional): Generate schemas using aliases defined. Defaults to True.
            namespace (Optional[str], optional): Optional namespace string for schema generation.

        Returns:
            Dict[str, Any]: The Avro schema for the model.
        """

        schema = pydantic_model.model_json_schema(by_alias=by_alias)

        if namespace is None:
            # default namespace will be based on title
            namespace = schema["title"]

        return cls._avro_schema(schema, namespace)

    @classmethod
    def avro_schema(
        cls, by_alias: bool = True, namespace: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Returns the Avro schema for the Pydantic class.

        Args:
            by_alias (bool, optional): Generate schemas using aliases defined. Defaults to True.
            namespace (Optional[str], optional): Optional namespace string for schema generation.

        Returns:
            Dict[str, Any]: The Avro schema for the model.
        """
        schema = cls.schema(by_alias=by_alias)

        if namespace is None:
            # default namespace will be based on title
            namespace = schema["title"]

        return cls._avro_schema(schema, namespace)

    @staticmethod
    def _avro_schema(schema: Dict[str, Any], namespace: str) -> Dict[str, Any]:
        """Return the avro schema for the given pydantic schema"""

        classes_seen = set()

        def get_definition(ref: str, schema: Dict[str, Any]) -> Dict[str, Any]:
            """Reading definition of base schema for nested structs"""
            id = ref.replace("#/definitions/", "")
            d = schema.get("definitions", {}).get(id)
            if d is None:
                raise RuntimeError(f"Definition {id} does not exist")
            return d  # type: ignore

        def get_type(value: Dict[str, Any]) -> Dict[str, Any]:
            """Returns a type of a single field"""
            t = value.get("type")
            f = value.get("format")
            r = value.get("$ref")
            a = value.get("additionalProperties")
            avro_type_dict: Dict[str, Any] = {}
            if "default" in value:
                avro_type_dict["default"] = value.get("default")
            if "description" in value:
                avro_type_dict["doc"] = value.get("description")
            if "allOf" in value and len(value["allOf"]) == 1:
                r = value["allOf"][0]["$ref"]
            if r is not None:
                class_name = r.replace("#/definitions/", "")
                if class_name in classes_seen:
                    avro_type_dict["type"] = class_name
                else:
                    d = get_definition(r, schema)
                    if "enum" in d:
                        avro_type_dict["type"] = {
                            "type": "enum",
                            "symbols": [str(v) for v in d["enum"]],
                            "name": d["title"],
                        }
                    else:
                        avro_type_dict["type"] = {
                            "type": "record",
                            "fields": get_fields(d),
                            # Name of the struct should be unique true the complete schema
                            # Because of this the path in the schema is tracked and used as name for a nested struct/array
                            "name": class_name,
                        }
                    classes_seen.add(class_name)
            elif t == "array":
                items = value.get("items")
                tn = get_type(items)  # type: ignore
                # If items in array are a object:
                if "$ref" in items:  # type: ignore
                    tn = tn["type"]
                # If items in array are a logicalType
                if (
                    isinstance(tn, dict)
                    and isinstance(tn.get("type", {}), dict)
                    and tn.get("type", {}).get("logicalType") is not None
                ):
                    tn = tn["type"]
                avro_type_dict["type"] = {"type": "array", "items": tn}
            elif t == "string" and f == "date-time":
                avro_type_dict["type"] = {
                    "type": "long",
                    "logicalType": "timestamp-micros",
                }
            elif t == "string" and f == "date":
                avro_type_dict["type"] = {
                    "type": "int",
                    "logicalType": "date",
                }
            elif t == "string" and f == "time":
                avro_type_dict["type"] = {
                    "type": "long",
                    "logicalType": "time-micros",
                }
            elif t == "string" and f == "uuid":
                avro_type_dict["type"] = {
                    "type": "string",
                    "logicalType": "uuid",
                }
            elif t == "string":
                avro_type_dict["type"] = "string"
            elif t == "number":
                avro_type_dict["type"] = "double"
            elif t == "integer":
                # integer in python can be a long
                avro_type_dict["type"] = "long"
            elif t == "boolean":
                avro_type_dict["type"] = "boolean"
            elif t == "object":
                if a is None:
                    value_type = "string"
                else:
                    value_type = get_type(a)  # type: ignore
                if isinstance(value_type, dict) and len(value_type) == 1:
                    value_type = value_type.get("type")  # type: ignore
                avro_type_dict["type"] = {"type": "map", "values": value_type}
            else:
                raise NotImplementedError(
                    f"Type '{t}' not support yet, "
                    f"please report this at https://github.com/godatadriven/pydantic-avro/issues"
                )
            return avro_type_dict

        def get_fields(s: Dict[str, Any]) -> List[Dict[str, Any]]:
            """Return a list of fields of a struct"""
            fields = []

            required = s.get("required", [])
            for key, value in s.get("properties", {}).items():
                if "type" not in value and "anyOf" in value:
                    any_of_types = value.pop("anyOf")
                    types = [x["type"] for x in any_of_types if x["type"] != "null"]
                    value["type"] = types[0]
                avro_type_dict = get_type(value)
                avro_type_dict["name"] = key

                if key not in required:
                    if avro_type_dict.get("default") is None:
                        avro_type_dict["type"] = ["null", avro_type_dict["type"]]
                        avro_type_dict["default"] = None

                fields.append(avro_type_dict)
            return fields

        fields = get_fields(schema)

        return {
            "type": "record",
            "namespace": namespace,
            "name": schema["title"],
            "fields": fields,
        }

# %% ../../../nbs/018_Avro_Encode_Decoder.ipynb 11
@export("fastkafka.encoder")
def avro_encoder(msg: BaseModel) -> bytes:
    """
    Encoder to encode pydantic instances to avro message

    Args:
        msg: An instance of pydantic basemodel

    Returns:
        A bytes message which is encoded from pydantic basemodel
    """
    schema = fastavro.schema.parse_schema(AvroBase.avro_schema_for_pydantic_object(msg))
    bytes_writer = io.BytesIO()

    d = msg.model_dump()
    for k, v in d.items():
        if "pydantic_core" in str(type(v)):
            d[k] = str(v)

    fastavro.schemaless_writer(bytes_writer, schema, d)
    raw_bytes = bytes_writer.getvalue()
    return raw_bytes

# %% ../../../nbs/018_Avro_Encode_Decoder.ipynb 13
@export("fastkafka.encoder")
def avro_decoder(raw_msg: bytes, cls: Type[BaseModel]) -> Any:
    """
    Decoder to decode avro encoded messages to pydantic model instance

    Args:
        raw_msg: Avro encoded bytes message received from Kafka topic
        cls: Pydantic class; This pydantic class will be used to construct instance of same class

    Returns:
        An instance of given pydantic class
    """
    schema = fastavro.schema.parse_schema(AvroBase.avro_schema_for_pydantic_class(cls))

    bytes_reader = io.BytesIO(raw_msg)
    msg_dict = fastavro.schemaless_reader(bytes_reader, schema)

    return cls(**msg_dict)

# %% ../../../nbs/018_Avro_Encode_Decoder.ipynb 16
@export("fastkafka.encoder")
def avsc_to_pydantic(schema: Dict[str, Any]) -> Type[BaseModel]:
    """
    Generate pydantic model from given Avro Schema

    Args:
        schema: Avro schema in dictionary format

    Returns:
        Pydantic model class built from given avro schema
    """
    if "type" not in schema or schema["type"] != "record":
        raise AttributeError("Type not supported")
    if "name" not in schema:
        raise AttributeError("Name is required")
    if "fields" not in schema:
        raise AttributeError("fields are required")

    classes = {}

    def get_python_type(t: Union[str, Dict[str, Any]]) -> str:
        """Returns python type for given avro type"""
        optional = False
        if isinstance(t, str):
            if t == "string":
                py_type = "str"
            elif t == "long" or t == "int":
                py_type = "int"
            elif t == "boolean":
                py_type = "bool"
            elif t == "double" or t == "float":
                py_type = "float"
            elif t in classes:
                py_type = t
            else:
                raise NotImplementedError(f"Type {t} not supported yet")
        elif isinstance(t, list):
            if "null" in t:
                optional = True
            if len(t) > 2 or (not optional and len(t) > 1):
                raise NotImplementedError("Only a single type ia supported yet")
            c = t.copy()
            c.remove("null")
            py_type = get_python_type(c[0])
        elif t.get("logicalType") == "uuid":
            py_type = "UUID"
        elif t.get("logicalType") == "decimal":
            py_type = "Decimal"
        elif (
            t.get("logicalType") == "timestamp-millis"
            or t.get("logicalType") == "timestamp-micros"
        ):
            py_type = "datetime"
        elif (
            t.get("logicalType") == "time-millis"
            or t.get("logicalType") == "time-micros"
        ):
            py_type = "time"
        elif t.get("logicalType") == "date":
            py_type = "date"
        elif t.get("type") == "enum":
            enum_name = t.get("name")
            if enum_name not in classes:
                enum_class = f"class {enum_name}(str, Enum):\n"
                for s in t.get("symbols"):  # type: ignore
                    enum_class += f'    {s} = "{s}"\n'
                classes[enum_name] = enum_class
            py_type = enum_name  # type: ignore
        elif t.get("type") == "string":
            py_type = "str"
        elif t.get("type") == "array":
            sub_type = get_python_type(t.get("items"))  # type: ignore
            py_type = f"List[{sub_type}]"
        elif t.get("type") == "record":
            record_type_to_pydantic(t)
            py_type = t.get("name")  # type: ignore
        elif t.get("type") == "map":
            value_type = get_python_type(t.get("values"))  # type: ignore
            py_type = f"Dict[str, {value_type}]"
        else:
            raise NotImplementedError(
                f"Type {t} not supported yet, "
                f"please report this at https://github.com/godatadriven/pydantic-avro/issues"
            )
        if optional:
            return f"Optional[{py_type}]"
        else:
            return py_type

    def record_type_to_pydantic(schema: Dict[str, Any]) -> Type[BaseModel]:
        """Convert a single avro record type to a pydantic class"""
        name = (
            schema["name"]
            if "." not in schema["name"]
            else schema["name"].split(".")[-1]
        )
        current = f"class {schema['name']}(BaseModel):\n"

        kwargs: Dict[str, Tuple[str, Any]] = {}

        if len(schema["fields"]) == 0:
            raise ValueError("Avro schema has no fields")

        for field in schema["fields"]:
            n = field["name"]
            t = get_python_type(field["type"])
            default = field.get("default")
            if "default" not in field:
                kwargs[n] = (t, ...)
                current += f"    {n}: {t}\n"
            elif isinstance(default, (bool, type(None))):
                kwargs[n] = (t, default)
                current += f"    {n}: {t} = {default}\n"
            else:
                kwargs[n] = (t, default)
                current += f"    {n}: {t} = {json.dumps(default)}\n"

        classes[name] = current
        pydantic_model = create_model(name, __module__=__name__, **kwargs)  # type: ignore
        return pydantic_model  # type: ignore

    return record_type_to_pydantic(schema)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/001_InMemoryBroker.ipynb.

# %% auto 0
__all__ = ['logger', 'KafkaRecord', 'KafkaPartition', 'KafkaTopic', 'split_list', 'GroupMetadata', 'InMemoryBroker',
           'InMemoryConsumer', 'InMemoryProducer', 'MockBatch']

# %% ../../nbs/001_InMemoryBroker.ipynb 1
import asyncio
import hashlib
import random
import string
import uuid
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import *
from types import ModuleType

from aiokafka.structs import ConsumerRecord, RecordMetadata, TopicPartition

import fastkafka._components.aiokafka_consumer_loop
import fastkafka._aiokafka_imports
from .._aiokafka_imports import AIOKafkaConsumer, AIOKafkaProducer
from .._components.logger import get_logger
from fastkafka._components.meta import (
    _get_default_kwargs_from_sig,
    classcontextmanager,
    delegates,
    patch,
)

# %% ../../nbs/001_InMemoryBroker.ipynb 3
logger = get_logger(__name__)

# %% ../../nbs/001_InMemoryBroker.ipynb 6
@dataclass
class KafkaRecord:
    topic: str = ""
    partition: int = 0
    key: Optional[bytes] = None
    value: bytes = b""
    offset: int = 0
    timestamp = 0
    timestamp_type = 0
    checksum = 0
    serialized_key_size = 0
    serialized_value_size = 0
    headers: Sequence[Tuple[str, bytes]] = field(default_factory=list)

# %% ../../nbs/001_InMemoryBroker.ipynb 7
class KafkaPartition:
    def __init__(self, *, partition: int, topic: str):
        """
        Initialize a KafkaPartition object.

        Args:
            partition: The partition number.
            topic: The topic name.
        """
        self.partition = partition
        self.topic = topic
        self.messages: List[KafkaRecord] = list()

    def write(self, value: bytes, key: Optional[bytes] = None) -> RecordMetadata:  # type: ignore
        """
        Write a Kafka record to the partition.

        Args:
            value: The value of the record.
            key: The key of the record.

        Returns:
            The record metadata.
        """
        record = KafkaRecord(
            topic=self.topic,
            partition=self.partition,
            value=value,
            key=key,
            offset=len(self.messages),
        )
        record_meta = RecordMetadata(
            topic=self.topic,
            partition=self.partition,
            topic_partition=TopicPartition(topic=self.topic, partition=self.partition),
            offset=len(self.messages),
            timestamp=1680602752070,
            timestamp_type=0,
            log_start_offset=0,
        )
        self.messages.append(record)
        return record_meta

    def read(self, offset: int) -> Tuple[List[KafkaRecord], int]:
        """
        Read Kafka records from the partition starting from the given offset.

        Args:
            offset: The starting offset.

        Returns:
            A tuple containing the list of records and the current offset.
        """
        return self.messages[offset:], len(self.messages)

    def latest_offset(self) -> int:
        """
        Get the latest offset of the partition.

        Returns:
            The latest offset.
        """
        return len(self.messages)

# %% ../../nbs/001_InMemoryBroker.ipynb 11
class KafkaTopic:
    def __init__(self, topic: str, num_partitions: int = 1):
        """
        Initialize a KafkaTopic object.

        Args:
            topic: The topic name.
            num_partitions: The number of partitions in the topic (default: 1).
        """
        self.topic = topic
        self.num_partitions = num_partitions
        self.partitions: List[KafkaPartition] = [
            KafkaPartition(topic=topic, partition=partition_index)
            for partition_index in range(num_partitions)
        ]

    def read(  # type: ignore
        self, partition: int, offset: int
    ) -> Tuple[TopicPartition, List[KafkaRecord], int]:
        """
        Read records from the specified partition and offset.

        Args:
            partition: The partition index.
            offset: The offset from which to start reading.

        Returns:
            A tuple containing the topic partition, list of Kafka records, and the new offset.
        """
        topic_partition = TopicPartition(topic=self.topic, partition=partition)
        records, offset = self.partitions[partition].read(offset)
        return topic_partition, records, offset

    def write_with_partition(  # type: ignore
        self,
        value: bytes,
        partition: int,
    ) -> RecordMetadata:
        """
        Write a record with a specified partition.

        Args:
            value: The value of the record.
            partition: The partition to write the record to.

        Returns:
            The metadata of the written record.
        """
        return self.partitions[partition].write(value)

    def write_with_key(self, value: bytes, key: bytes) -> RecordMetadata:  # type: ignore
        """
        Write a record with a specified key.

        Args:
            value: The value of the record.
            key: The key of the record.

        Returns:
            The metadata of the written record.
        """
        partition = int(hashlib.sha256(key).hexdigest(), 16) % self.num_partitions
        return self.partitions[partition].write(value, key=key)

    def write(  # type: ignore
        self,
        value: bytes,
        *,
        key: Optional[bytes] = None,
        partition: Optional[int] = None,
    ) -> RecordMetadata:
        """
        Write a record to the topic.

        Args:
            value: The value of the record.
            key: The key of the record (optional).
            partition: The partition to write the record to (optional).

        Returns:
            The metadata of the written record.
        """
        if partition is not None:
            return self.write_with_partition(value, partition)

        if key is not None:
            return self.write_with_key(value, key)

        partition = random.randint(0, self.num_partitions - 1)  # nosec
        return self.write_with_partition(value, partition)

    def latest_offset(self, partition: int) -> int:
        """
        Get the latest offset of a partition.

        Args:
            partition: The partition index.

        Returns:
            The latest offset of the partition.
        """
        return self.partitions[partition].latest_offset()

# %% ../../nbs/001_InMemoryBroker.ipynb 17
def split_list(list_to_split: List[Any], split_size: int) -> List[List[Any]]:
    """
    Split a list into smaller lists of a specified size.

    Args:
        list_to_split: The list to split.
        split_size: The size of each split.

    Returns:
        A list of smaller lists.
    """
    return [
        list_to_split[start_index : start_index + split_size]
        for start_index in range(0, len(list_to_split), split_size)
    ]

# %% ../../nbs/001_InMemoryBroker.ipynb 19
class GroupMetadata:
    def __init__(self, num_partitions: int):
        """
        Initialize a GroupMetadata object.

        Args:
            num_partitions: The number of partitions in the group.
        """
        self.num_partitions = num_partitions
        self.partitions_offsets: Dict[int, int] = {}
        self.consumer_ids: List[uuid.UUID] = list()
        self.partition_assignments: Dict[uuid.UUID, List[int]] = {}

    def subscribe(self, consumer_id: uuid.UUID) -> None:
        """
        Subscribe a consumer to the group.

        Args:
            consumer_id: The ID of the consumer.
        """
        self.consumer_ids.append(consumer_id)
        self.rebalance()

    def unsubscribe(self, consumer_id: uuid.UUID) -> None:
        """
        Unsubscribe a consumer from the group.

        Args:
            consumer_id: The ID of the consumer.
        """
        self.consumer_ids.remove(consumer_id)
        self.rebalance()

    def rebalance(self) -> None:
        """
        Rebalance the group's partition assignments.
        """
        if len(self.consumer_ids) == 0:
            self.partition_assignments = {}
        else:
            partitions_per_actor = self.num_partitions // len(self.consumer_ids)
            if self.num_partitions % len(self.consumer_ids) != 0:
                partitions_per_actor += 1
            self.assign_partitions(partitions_per_actor)

    def assign_partitions(self, partitions_per_actor: int) -> None:
        partitions = [i for i in range(self.num_partitions)]

        partitions_split = split_list(partitions, partitions_per_actor)
        self.partition_assignments = {
            self.consumer_ids[i]: partition_split
            for i, partition_split in enumerate(partitions_split)
        }

    def get_partitions(
        self, consumer_id: uuid.UUID
    ) -> Tuple[List[int], Dict[int, Optional[int]]]:
        """
        Get the partition assignments and offsets for a consumer.

        Args:
            consumer_id: The ID of the consumer.

        Returns:
            A tuple containing the partition assignments and offsets.
        """
        partition_assignments = self.partition_assignments.get(consumer_id, [])
        partition_offsets_assignments = {
            partition: self.partitions_offsets.get(partition, None)
            for partition in partition_assignments
        }
        return partition_assignments, partition_offsets_assignments

    def set_offset(self, partition: int, offset: int) -> None:
        """
        Set the offset for a partition.

        Args:
            partition: The partition index.
            offset: The offset to set.
        """
        self.partitions_offsets[partition] = offset

# %% ../../nbs/001_InMemoryBroker.ipynb 22
@classcontextmanager()
class InMemoryBroker:
    def __init__(
        self,
        num_partitions: int = 1,
    ):
        self.num_partitions = num_partitions
        self.topics: Dict[Tuple[str, str], KafkaTopic] = {}
        self.topic_groups: Dict[Tuple[str, str, str], GroupMetadata] = {}
        self.is_started: bool = False

    def connect(self) -> uuid.UUID:
        return uuid.uuid4()

    def dissconnect(self, consumer_id: uuid.UUID) -> None:
        """
        Disconnect a consumer from the broker.

        Args:
            consumer_id: The ID of the consumer.
        """
        pass

    def subscribe(
        self, bootstrap_server: str, topic: str, group: str, consumer_id: uuid.UUID
    ) -> None:
        raise NotImplementedError()

    def unsubscribe(
        self, bootstrap_server: str, topic: str, group: str, consumer_id: uuid.UUID
    ) -> None:
        raise NotImplementedError()

    def read(  # type: ignore
        self,
        *,
        bootstrap_server: str,
        topic: str,
        group: str,
        consumer_id: uuid.UUID,
        auto_offset_reset: str,
    ) -> Dict[TopicPartition, List[KafkaRecord]]:
        raise NotImplementedError()

    def write(  # type: ignore
        self,
        *,
        bootstrap_server: str,
        topic: str,
        value: bytes,
        key: Optional[bytes] = None,
        partition: Optional[int] = None,
    ) -> RecordMetadata:
        raise NotImplementedError()

    @contextmanager
    def lifecycle(self) -> Iterator["InMemoryBroker"]:
        """
        Context manager for the lifecycle of the in-memory broker.

        Yields:
            An instance of the in-memory broker.
        """
        raise NotImplementedError()

    async def _start(self) -> str:
        """
        Start the in-memory broker.

        Returns:
            The address of the broker.
        """
        logger.info("InMemoryBroker._start() called")
        self.__enter__()  # type: ignore
        return "localbroker:0"

    async def _stop(self) -> None:
        """
        Stop the in-memory broker.
        """
        logger.info("InMemoryBroker._stop() called")
        self.__exit__(None, None, None)  # type: ignore

# %% ../../nbs/001_InMemoryBroker.ipynb 23
@patch
def subscribe(
    self: InMemoryBroker,
    bootstrap_server: str,
    topic: str,
    group: str,
    consumer_id: uuid.UUID,
) -> None:
    """
    Subscribe a consumer to a topic group.

    Args:
        bootstrap_server: The bootstrap server address.
        topic: The topic to subscribe to.
        group: The group to join.
        consumer_id: The ID of the consumer.
    """
    if (bootstrap_server, topic) not in self.topics:
        self.topics[(bootstrap_server, topic)] = KafkaTopic(
            topic=topic, num_partitions=self.num_partitions
        )

    group_meta = self.topic_groups.get(
        (bootstrap_server, topic, group), GroupMetadata(self.num_partitions)
    )
    group_meta.subscribe(consumer_id)
    self.topic_groups[(bootstrap_server, topic, group)] = group_meta


@patch
def unsubscribe(
    self: InMemoryBroker,
    bootstrap_server: str,
    topic: str,
    group: str,
    consumer_id: uuid.UUID,
) -> None:
    """
    Unsubscribe a consumer from a topic group.

    Args:
        bootstrap_server: The bootstrap server address.
        topic: The topic to unsubscribe from.
        group: The group to leave.
        consumer_id: The ID of the consumer.
    """
    self.topic_groups[(bootstrap_server, topic, group)].unsubscribe(consumer_id)

# %% ../../nbs/001_InMemoryBroker.ipynb 25
@patch
def write(  # type: ignore
    self: InMemoryBroker,
    *,
    bootstrap_server: str,
    topic: str,
    value: bytes,
    key: Optional[bytes] = None,
    partition: Optional[int] = None,
) -> RecordMetadata:
    """
    Write a message to a topic.

    Args:
        bootstrap_server: The bootstrap server address.
        topic: The topic to write the message to.
        value: The value of the message.
        key: The key associated with the message.
        partition: The partition ID to write the message to.

    Returns:
        The metadata of the written message.
    """
    if (bootstrap_server, topic) not in self.topics:
        self.topics[(bootstrap_server, topic)] = KafkaTopic(
            topic=topic, num_partitions=self.num_partitions
        )

    return self.topics[(bootstrap_server, topic)].write(
        value, key=key, partition=partition
    )

# %% ../../nbs/001_InMemoryBroker.ipynb 27
@patch
def read(  # type: ignore
    self: InMemoryBroker,
    *,
    bootstrap_server: str,
    topic: str,
    group: str,
    consumer_id: uuid.UUID,
    auto_offset_reset: str,
) -> Dict[TopicPartition, List[KafkaRecord]]:
    """
    Read messages from a topic group.

    Args:
        bootstrap_server: The bootstrap server address.
        topic: The topic to read messages from.
        group: The group to read messages for.
        consumer_id: The ID of the consumer.
        auto_offset_reset: The strategy to use when the consumer does not have a valid offset for the group.

    Returns:
        A dictionary containing the messages retrieved from each topic partition.
    """
    group_meta = self.topic_groups[(bootstrap_server, topic, group)]
    partitions, offsets = group_meta.get_partitions(consumer_id)

    if len(partitions) == 0:
        return {}

    partitions_data = {}

    for partition in partitions:
        offset = offsets[partition]

        if offset is None:
            offset = (
                self.topics[(bootstrap_server, topic)].latest_offset(partition)
                if auto_offset_reset == "latest"
                else 0
            )

        topic_partition, data, offset = self.topics[(bootstrap_server, topic)].read(
            partition, offset
        )

        partitions_data[topic_partition] = data
        group_meta.set_offset(partition, offset)

    return partitions_data

# %% ../../nbs/001_InMemoryBroker.ipynb 34
# InMemoryConsumer
class InMemoryConsumer:
    def __init__(
        self,
        broker: InMemoryBroker,
    ) -> None:
        self.broker = broker
        self._id: Optional[uuid.UUID] = None
        self._auto_offset_reset: str = "latest"
        self._group_id: Optional[str] = None
        self._topics: List[str] = list()
        self._bootstrap_servers = ""

    @delegates(AIOKafkaConsumer)
    def __call__(self, **kwargs: Any) -> "InMemoryConsumer":
        defaults = _get_default_kwargs_from_sig(InMemoryConsumer.__call__, **kwargs)
        consume_copy = InMemoryConsumer(self.broker)
        consume_copy._auto_offset_reset = defaults["auto_offset_reset"]
        consume_copy._bootstrap_servers = (
            "".join(defaults["bootstrap_servers"])
            if isinstance(defaults["bootstrap_servers"], list)
            else defaults["bootstrap_servers"]
        )

        consume_copy._group_id = (
            defaults["group_id"]
            if defaults["group_id"] is not None
            else "".join(random.choices(string.ascii_letters, k=10))  # nosec
        )
        return consume_copy

    @delegates(AIOKafkaConsumer.start)
    async def start(self, **kwargs: Any) -> None:
        pass

    @delegates(AIOKafkaConsumer.stop)
    async def stop(self, **kwargs: Any) -> None:
        pass

    @delegates(AIOKafkaConsumer.subscribe)
    def subscribe(self, topics: List[str], **kwargs: Any) -> None:
        raise NotImplementedError()

    @delegates(AIOKafkaConsumer.getmany)
    async def getmany(  # type: ignore
        self, **kwargs: Any
    ) -> Dict[TopicPartition, List[ConsumerRecord]]:
        raise NotImplementedError()

# %% ../../nbs/001_InMemoryBroker.ipynb 37
@patch
@delegates(AIOKafkaConsumer.start)
async def start(self: InMemoryConsumer, **kwargs: Any) -> None:
    """
    Start consuming messages from the connected broker.

    Raises:
        RuntimeError: If start() has already been called without calling stop() first.
    """
    logger.info("AIOKafkaConsumer patched start() called()")
    if self._id is not None:
        raise RuntimeError(
            "Consumer start() already called! Run consumer stop() before running start() again"
        )
    self._id = self.broker.connect()

# %% ../../nbs/001_InMemoryBroker.ipynb 40
@patch  # type: ignore
@delegates(AIOKafkaConsumer.subscribe)
def subscribe(self: InMemoryConsumer, topics: List[str], **kwargs: Any) -> None:
    """
    Subscribe to a list of topics for consuming messages.

    Args:
        topics: A list of topics to subscribe to.

    Raises:
        RuntimeError: If start() has not been called before calling subscribe().
    """
    logger.info("AIOKafkaConsumer patched subscribe() called")
    if self._id is None:
        raise RuntimeError("Consumer start() not called! Run consumer start() first")
    logger.info(f"AIOKafkaConsumer.subscribe(), subscribing to: {topics}")
    for topic in topics:
        self.broker.subscribe(
            bootstrap_server=self._bootstrap_servers,
            consumer_id=self._id,
            topic=topic,
            group=self._group_id,  # type: ignore
        )
        self._topics.append(topic)

# %% ../../nbs/001_InMemoryBroker.ipynb 43
@patch
@delegates(AIOKafkaConsumer.stop)
async def stop(self: InMemoryConsumer, **kwargs: Any) -> None:
    """
    Stop consuming messages from the connected broker.

    Raises:
        RuntimeError: If start() has not been called before calling stop().
    """
    logger.info("AIOKafkaConsumer patched stop() called")
    if self._id is None:
        raise RuntimeError("Consumer start() not called! Run consumer start() first")
    for topic in self._topics:
        self.broker.unsubscribe(
            bootstrap_server=self._bootstrap_servers,
            topic=topic,
            group=self._group_id,  # type: ignore
            consumer_id=self._id,
        )

# %% ../../nbs/001_InMemoryBroker.ipynb 46
@patch
@delegates(AIOKafkaConsumer.getmany)
async def getmany(  # type: ignore
    self: InMemoryConsumer, **kwargs: Any
) -> Dict[TopicPartition, List[ConsumerRecord]]:
    """
    Retrieve messages from the subscribed topics.

    Returns:
        A dictionary containing the retrieved messages from each topic partition.

    Raises:
        RuntimeError: If start() has not been called before calling getmany().
    """
    await asyncio.sleep(0)
    for topic in self._topics:
        return self.broker.read(
            bootstrap_server=self._bootstrap_servers,
            topic=topic,
            consumer_id=self._id,  # type: ignore
            group=self._group_id,  # type: ignore
            auto_offset_reset=self._auto_offset_reset,
        )

# %% ../../nbs/001_InMemoryBroker.ipynb 49
class InMemoryProducer:
    def __init__(self, broker: InMemoryBroker, **kwargs: Any) -> None:
        self.broker = broker
        self.id: Optional[uuid.UUID] = None
        self._bootstrap_servers = ""

    @delegates(AIOKafkaProducer)
    def __call__(self, **kwargs: Any) -> "InMemoryProducer":
        defaults = _get_default_kwargs_from_sig(InMemoryConsumer.__call__, **kwargs)
        producer_copy = InMemoryProducer(self.broker)
        producer_copy._bootstrap_servers = (
            "".join(defaults["bootstrap_servers"])
            if isinstance(defaults["bootstrap_servers"], list)
            else defaults["bootstrap_servers"]
        )
        return producer_copy

    @delegates(AIOKafkaProducer.start)
    async def start(self, **kwargs: Any) -> None:
        raise NotImplementedError()

    @delegates(AIOKafkaProducer.stop)
    async def stop(self, **kwargs: Any) -> None:
        raise NotImplementedError()

    @delegates(AIOKafkaProducer.send)
    async def send(  # type: ignore
        self,
        topic: str,
        msg: bytes,
        key: Optional[bytes] = None,
        **kwargs: Any,
    ):
        raise NotImplementedError()

    @delegates(AIOKafkaProducer.partitions_for)
    async def partitions_for(self, topic: str) -> List[int]:
        raise NotImplementedError()

    @delegates(AIOKafkaProducer._partition)
    def _partition(
        self, topic: str, arg1: Any, arg2: Any, arg3: Any, key: bytes, arg4: Any
    ) -> int:
        raise NotImplementedError()

    @delegates(AIOKafkaProducer.create_batch)
    def create_batch(self) -> "MockBatch":
        raise NotImplementedError()

    @delegates(AIOKafkaProducer.send_batch)
    async def send_batch(self, batch: "MockBatch", topic: str, partition: Any) -> None:
        raise NotImplementedError()

# %% ../../nbs/001_InMemoryBroker.ipynb 52
@patch  # type: ignore
@delegates(AIOKafkaProducer.start)
async def start(self: InMemoryProducer, **kwargs: Any) -> None:
    """
    Start the in-memory producer.

    Raises:
        RuntimeError: If start() has already been called without calling stop() first.
    """
    logger.info("AIOKafkaProducer patched start() called()")
    if self.id is not None:
        raise RuntimeError(
            "Producer start() already called! Run producer stop() before running start() again"
        )
    self.id = self.broker.connect()

# %% ../../nbs/001_InMemoryBroker.ipynb 55
@patch  # type: ignore
@delegates(AIOKafkaProducer.stop)
async def stop(self: InMemoryProducer, **kwargs: Any) -> None:
    """
    Stop the in-memory producer.

    Raises:
        RuntimeError: If start() has not been called before calling stop().
    """
    logger.info("AIOKafkaProducer patched stop() called")
    if self.id is None:
        raise RuntimeError("Producer start() not called! Run producer start() first")

# %% ../../nbs/001_InMemoryBroker.ipynb 58
@patch
@delegates(AIOKafkaProducer.send)
async def send(  # type: ignore
    self: InMemoryProducer,
    topic: str,
    msg: bytes,
    key: Optional[bytes] = None,
    partition: Optional[int] = None,
    **kwargs: Any,
):  # asyncio.Task[RecordMetadata]
    """
    Send a message to the specified topic.

    Args:
        topic: The topic to send the message to.
        msg: The message to send.
        key: The key associated with the message (optional).
        partition: The partition to send the message to (optional).
        **kwargs: Additional arguments to be passed to AIOKafkaProducer.send().

    Returns:
        A task that resolves to the RecordMetadata of the sent message.

    Raises:
        RuntimeError: If start() has not been called before calling send().
    """
    if self.id is None:
        raise RuntimeError("Producer start() not called! Run producer start() first")

    record = self.broker.write(
        bootstrap_server=self._bootstrap_servers,
        topic=topic,
        value=msg,
        key=key,
        partition=partition,
    )

    async def _f(record: ConsumerRecord = record) -> RecordMetadata:  # type: ignore
        return record

    return asyncio.create_task(_f())

# %% ../../nbs/001_InMemoryBroker.ipynb 60
@patch
@delegates(AIOKafkaProducer.partitions_for)
async def partitions_for(self: InMemoryProducer, topic: str) -> List[int]:
    """
    Retrieve the list of partitions for the specified topic.

    Args:
        topic: The topic to get the partitions for.

    Returns:
        A list of partition IDs.
    """
    return [i for i in range(self.broker.num_partitions)]

# %% ../../nbs/001_InMemoryBroker.ipynb 62
@patch
@delegates(AIOKafkaProducer._partition)
def _partition(
    self: InMemoryProducer,
    topic: str,
    arg1: Any,
    arg2: Any,
    arg3: Any,
    key: bytes,
    arg4: Any,
) -> int:
    """
    Determine the partition to which the message should be sent.

    Args:
        topic: The topic to send the message to.
        arg1, arg2, arg3, arg4: Additional arguments passed to the original AIOKafkaProducer._partition().

    Returns:
        The partition ID.
    """
    return int(hashlib.sha256(key).hexdigest(), 16) % self.broker.num_partitions

# %% ../../nbs/001_InMemoryBroker.ipynb 64
class MockBatch:
    def __init__(self) -> None:
        """
        Initialize an instance of MockBatch.
        """
        self._batch: List[Tuple] = list()

    def append(  # type: ignore
        self, key: Optional[bytes], value: bytes, timestamp: int
    ) -> RecordMetadata:
        """
        Append a message to the batch.

        Args:
            key: The key associated with the message (optional).
            value: The value of the message.
            timestamp: The timestamp of the message.

        Returns:
            The RecordMetadata of the appended message.
        """
        self._batch.append((key, value))
        return RecordMetadata(
            topic="",
            partition=0,
            topic_partition=None,
            offset=0,
            timestamp=timestamp,
            timestamp_type=0,
            log_start_offset=0,
        )


@patch
@delegates(AIOKafkaProducer.create_batch)
def create_batch(self: InMemoryProducer) -> "MockBatch":
    """
    Create a mock batch for the in-memory producer.

    Returns:
        A MockBatch instance.
    """
    return MockBatch()


@patch
@delegates(AIOKafkaProducer.send_batch)
async def send_batch(
    self: InMemoryProducer, batch: "MockBatch", topic: str, partition: Any
) -> None:
    """
    Send a batch of messages to the specified topic and partition.

    Args:
        batch: The MockBatch containing the messages to send.
        topic: The topic to send the batch of messages to.
        partition: The partition to send the batch of messages to.
    """
    for record in batch._batch:
        self.broker.write(
            bootstrap_server=self._bootstrap_servers,
            topic=topic,
            value=record[1],
            key=record[0],
            partition=partition,
        )

# %% ../../nbs/001_InMemoryBroker.ipynb 68
@patch
@contextmanager
def lifecycle(self: InMemoryBroker) -> Iterator[InMemoryBroker]:
    """
    Context manager for the lifecycle of the in-memory broker.

    Yields:
        An instance of the in-memory broker.
    """
    logger.info(
        "InMemoryBroker._patch_consumers_and_producers(): Patching consumers and producers!"
    )
    try:
        logger.info("InMemoryBroker starting")

        old_consumer = fastkafka._aiokafka_imports.AIOKafkaConsumer
        old_producer = fastkafka._aiokafka_imports.AIOKafkaProducer

        fastkafka._aiokafka_imports.AIOKafkaConsumer = InMemoryConsumer(self)
        fastkafka._aiokafka_imports.AIOKafkaProducer = InMemoryProducer(self)

        self.is_started = True
        yield self
    finally:
        logger.info("InMemoryBroker stopping")

        fastkafka._aiokafka_imports.AIOKafkaConsumer = old_consumer
        fastkafka._aiokafka_imports.AIOKafkaProducer = old_producer

        self.is_started = False



__version__ = "0.8.0"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/010_Application_export.ipynb.

# %% auto 0
__all__ = ['dummy']

# %% ../nbs/010_Application_export.ipynb 1
from ._application.app import FastKafka
from ._components.meta import export
from ._components.producer_decorator import KafkaEvent
from ._components.aiokafka_consumer_loop import EventMetadata

__all__ = ["FastKafka", "KafkaEvent", "EventMetadata"]

# %% ../nbs/010_Application_export.ipynb 2
@export("_dummy")
def dummy() -> None:
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/020_Encoder_Export.ipynb.

# %% auto 0
__all__ = ['dummy']

# %% ../nbs/020_Encoder_Export.ipynb 1
from fastkafka._components.encoder.avro import (
    AvroBase,
    avro_decoder,
    avro_encoder,
    avsc_to_pydantic,
)
from ._components.encoder.json import json_decoder, json_encoder
from ._components.meta import export

__all__ = [
    "AvroBase",
    "avro_decoder",
    "avro_encoder",
    "avsc_to_pydantic",
    "json_decoder",
    "json_encoder",
]

# %% ../nbs/020_Encoder_Export.ipynb 3
@export("_dummy")
def dummy() -> None:
    pass
